(*
 * Copyright (C) 2013, National ICT Australia Limited. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *  * The name of National ICT Australia Limited nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *)

theory WordAbstract
imports L2Defs
begin

definition "INT_MAX \<equiv> (2 :: int) ^ 31 - 1"
definition "INT_MIN \<equiv> - ((2 :: int) ^ 31)"
definition "UINT_MAX \<equiv> (2 :: nat) ^ 32 - 1"

lemma INT_MAX_value [code]: "INT_MAX \<equiv> 2147483647"
  by (clarsimp simp: INT_MAX_def)
lemma INT_MIN_value [code]: "INT_MIN \<equiv> -2147483648"
  by (clarsimp simp: INT_MIN_def)
lemma UINT_MAX_value [code]: "UINT_MAX \<equiv> 4294967295"
  by (clarsimp simp: UINT_MAX_def)

lemma INT_MIN_MAX_lemmas [simp]:
  "unat (u :: word32) \<le> UINT_MAX"
  "sint (s :: sword32) \<le> INT_MAX"
  "INT_MIN \<le> sint (s :: sword32)"
  "INT_MIN \<le> INT_MAX"
  "INT_MIN \<le> sint (s :: sword32)"
  "INT_MIN \<le> INT_MAX"
  "INT_MIN \<le> 0"
  "0 \<le> INT_MAX"
  unfolding UINT_MAX_def INT_MAX_def INT_MIN_def
  using sint_range_size [where w=s, simplified word_size, simplified]
        unat_lt2p [where 'a=32, simplified]
        zle_add1_eq_le [where z=INT_MAX, symmetric]
        less_eq_Suc_le
  by auto

(* Show that a binary operator "X" (of type "'a \<Rightarrow> 'a \<Rightarrow> bool") abstracts "X'". *)
definition
  abstract_bool_binop :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> ('c \<Rightarrow> 'a)
               \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> ('c \<Rightarrow> 'c \<Rightarrow> bool) \<Rightarrow> bool"
where
  "abstract_bool_binop P A X X' \<equiv> \<forall>a b. P (A a) (A b) \<longrightarrow> (X' a b = X (A a) (A b))"

(* Show that a binary operator "X" (of type "'a \<Rightarrow> 'a \<Rightarrow> 'b") abstracts "X'". *)
definition
  abstract_binop :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> ('c \<Rightarrow> 'a)
               \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> ('c \<Rightarrow> 'c \<Rightarrow> 'c) \<Rightarrow> bool"
where
   "abstract_binop P A X X' \<equiv> \<forall>a b. P (A a) (A b) \<longrightarrow> (A (X' a b) = X (A a) (A b))"

(* The value "a" is the abstract version of "b" under precondition "P". *)
definition "abstract_val P A a b \<equiv> P \<longrightarrow> (a = A b)"

declare abstract_bool_binop_def [simp]
declare abstract_binop_def [simp]
declare abstract_val_def [simp]

lemma abstract_binop_is_abstract_val:
    "abstract_binop P A X X' = (\<forall>a b. abstract_val (P (A a) (A b)) A (X (A a) (A b)) (X' a b))"
  by auto

lemma abstract_expr_bool_binop:
  "\<lbrakk> abstract_bool_binop E A X X';
     abstract_val P A a a';
     abstract_val Q A b b' \<rbrakk> \<Longrightarrow>
           abstract_val (P \<and> Q \<and> E a b) (\<lambda>x. x) (X a b) (X' a' b')"
  by clarsimp

lemma abstract_expr_binop:
  "\<lbrakk> abstract_binop E A X X';
     abstract_val P A a a';
     abstract_val Q A b b' \<rbrakk> \<Longrightarrow>
           abstract_val (P \<and> Q \<and> E a b) A (X a b) (X' a' b')"
  by clarsimp

lemma unat_abstract_bool_binops:
    "abstract_bool_binop (\<lambda>_ _. True) (unat :: word32 \<Rightarrow> nat) (op <) (op <)"
    "abstract_bool_binop (\<lambda>_ _. True) (unat :: word32 \<Rightarrow> nat) (op \<le>) (op \<le>)"
    "abstract_bool_binop (\<lambda>_ _. True) (unat :: word32 \<Rightarrow> nat) (op =) (op =)"
  by (auto simp:  word_less_nat_alt word_le_nat_alt eq_iff)

lemma unat_abstract_binops:
  "abstract_binop (\<lambda>a b. a + b \<le> UINT_MAX) (unat :: word32 \<Rightarrow> nat) (op +) (op +)"
  "abstract_binop (\<lambda>a b. a * b \<le> UINT_MAX) (unat :: word32 \<Rightarrow> nat) (op * ) (op * )"
  "abstract_binop (\<lambda>a b. a \<ge> b) (unat :: word32 \<Rightarrow> nat) (op -) (op -)"
  "abstract_binop (\<lambda>a b. True) (unat :: word32 \<Rightarrow> nat) (op div) (op div)"
  "abstract_binop (\<lambda>a b. True) (unat :: word32 \<Rightarrow> nat) (op mod) (op mod)"
  apply (auto simp: unat_plus_if' unat_div unat_mod UINT_MAX_def
              unat_mult_simple word_bits_def unat_sub word_le_nat_alt)
  done

lemma snat_abstract_bool_binops:
    "abstract_bool_binop (\<lambda>_ _. True) (sint :: sword32 \<Rightarrow> int) (op <) (word_sless)"
    "abstract_bool_binop (\<lambda>_ _. True) (sint :: sword32 \<Rightarrow> int) (op \<le>) (word_sle)"
    "abstract_bool_binop (\<lambda>_ _. True) (sint :: sword32 \<Rightarrow> int) (op =) (op =)"
  apply (auto simp: word_sless_def word_sle_def less_le)
  done

lemma snat_abstract_binops:
  "abstract_binop (\<lambda>a b. INT_MIN \<le> a + b \<and> a + b \<le> INT_MAX) (sint :: sword32 \<Rightarrow> int) (op +) (op +)"
  "abstract_binop (\<lambda>a b. INT_MIN \<le> a * b \<and> a * b \<le> INT_MAX) (sint :: sword32 \<Rightarrow> int) (op * ) (op * )"
  "abstract_binop (\<lambda>a b. INT_MIN \<le> a - b \<and> a - b \<le> INT_MAX) (sint :: sword32 \<Rightarrow> int) (op -) (op -)"
  "abstract_binop (\<lambda>a b. INT_MIN \<le> a sdiv b \<and> a sdiv b \<le> INT_MAX) (sint :: sword32 \<Rightarrow> int) (op sdiv) (op sdiv)"
  "abstract_binop (\<lambda>a b. INT_MIN \<le> a smod b \<and> a smod b \<le> INT_MAX) (sint :: sword32 \<Rightarrow> int) (op smod) (op smod)"
  apply (auto simp: signed_arith_sint word_size INT_MIN_def INT_MAX_def)
  done

lemma abstract_if:
  "\<lbrakk> abstract_val P1 (\<lambda>x. x) a' a;
     abstract_val P2 A b' b;
     abstract_val P3 A c' c \<rbrakk> \<Longrightarrow>
      abstract_val (P1 \<and> P2 \<and> P3) A (if a' then b' else c') (if a then b else c)"
  by clarsimp

lemmas word_abs_word_ops [simplified simp_thms] =

  abstract_expr_bool_binop [OF snat_abstract_bool_binops(1)]
  abstract_expr_bool_binop [OF snat_abstract_bool_binops(2)]
  abstract_expr_bool_binop [OF snat_abstract_bool_binops(3)]
  abstract_expr_binop [OF snat_abstract_binops(1)]
  abstract_expr_binop [OF snat_abstract_binops(2)]
  abstract_expr_binop [OF snat_abstract_binops(3)]
  abstract_expr_binop [OF snat_abstract_binops(4)]
  abstract_expr_binop [OF snat_abstract_binops(5)]

  abstract_expr_bool_binop [OF unat_abstract_bool_binops(1)]
  abstract_expr_bool_binop [OF unat_abstract_bool_binops(2)]
  abstract_expr_bool_binop [OF unat_abstract_bool_binops(3)]
  abstract_expr_binop [OF unat_abstract_binops(1)]
  abstract_expr_binop [OF unat_abstract_binops(2)]
  abstract_expr_binop [OF unat_abstract_binops(3)]
  abstract_expr_binop [OF unat_abstract_binops(4)]
  abstract_expr_binop [OF unat_abstract_binops(5)]

  abstract_if



lemma abstract_val_negate [word_abs]:
    "abstract_val P (\<lambda>x. x) C' C \<Longrightarrow>  abstract_val P (\<lambda>x. x) (\<not> C') (\<not> C)"
  by simp

(* FIXME : possibly too strict; perhaps allow overflow during the cast. *)
(*
lemma abstract_val_ucast [word_abs]:
    "abstract_val P unat C' C
            \<Longrightarrow>  abstract_val (P \<and> C' \<le> INT_MAX) sint (int C') (ucast (C :: word32) :: sword32)"
  apply (clarsimp simp: uint_nat [symmetric])
  apply (subst sint_eq_uint)
   apply (rule not_msb_from_less)
   apply (clarsimp simp: word_less_nat_alt unat_ucast)
  apply (clarsimp simp: uint_up_ucast is_up)
  done
*)

(* FIXME : possibly too strict; perhaps allow underflow during the cast. *)
(*
lemma abstract_val_scast [word_abs]:
    "abstract_val P sint C' C
            \<Longrightarrow>  abstract_val (P \<and> 0 \<le> C') unat (nat C') (scast (C :: sword32) :: word32)"
  apply (clarsimp simp: down_cast_same [symmetric] is_down unat_ucast)
  apply (subst sint_eq_uint)
   apply (clarsimp simp: word_msb_sint)
  apply (clarsimp simp: unat_def [symmetric])
  apply (subst word_unat.norm_Rep [symmetric])
  apply clarsimp
  done
*)

definition
 "valid_typ_abs_fn (P :: 'a \<Rightarrow> bool) (A :: 'c \<Rightarrow> 'a) (C :: 'a \<Rightarrow> 'c) \<equiv>
     (\<forall>v. P v \<longrightarrow> A (C v) = v)"


declare valid_typ_abs_fn_def [simp]

lemma valid_typ_abs_fn_id [word_abs]:
  "valid_typ_abs_fn \<top> (\<lambda>x. x) (\<lambda>x. x)"
  by clarsimp

lemma valid_typ_abs_fn_unit [word_abs]:
  "valid_typ_abs_fn \<top> (\<lambda>x. ()) (\<lambda>x. x :: unit)"
  by clarsimp

lemma valid_typ_abs_fn_unat [word_abs]:
  "valid_typ_abs_fn (\<lambda>v. v \<le> UINT_MAX) (unat :: word32 \<Rightarrow> nat) (of_nat :: nat \<Rightarrow> word32)"
  apply (clarsimp simp: unat_of_nat_eq UINT_MAX_def)
  done


definition "corresTA rx ex A C \<equiv> corresXF (\<lambda>s. s) (\<lambda>r s. rx r) (\<lambda>r s. ex r) \<top> A C"

definition "corres_rel' rr A C \<equiv>
    \<forall>s. \<not> snd (A s) \<longrightarrow>
        ((\<forall>(r', t') \<in> fst (C s). \<exists>(r, t) \<in> fst (A s). (t = t' \<and> (r, r') \<in> rr))
                \<and> \<not> snd (C s))"

definition "corres_rel ret_rr ex_rr A C \<equiv>
    \<forall>s. \<not> snd (A s) \<longrightarrow>
        ((\<forall>(r', t') \<in> fst (C s).
          case r' of
             Inl r'_val \<Rightarrow> \<exists>(r, t) \<in> fst (A s). (t = t' \<and> (\<exists>r_val. r = Inl r_val \<and> (r_val, r'_val) \<in> ex_rr))
           | Inr r'_val \<Rightarrow> \<exists>(r, t) \<in> fst (A s). (t = t' \<and> (\<exists>r_val. r = Inr r_val \<and> (r_val, r'_val) \<in> ret_rr))
        \<and> \<not> snd (C s)))"

lemma corres_rel'_return: "corres_rel' {(a, b). a = (b \<noteq> (0::word32))} (return (x \<noteq> 0)) (return (x :: word32))"
  apply (monad_eq simp: corres_rel'_def)
  done

lemma corres_rel'_seq: "
  \<lbrakk> corres_rel' {(a, b). a = (b \<noteq> (0::word32))} A A';
     \<And>x x'. (x = (x' \<noteq> 0)) \<Longrightarrow>  corres_rel' {(a, b). a = (b \<noteq> (0 :: word32))} (B x) (B' x') \<rbrakk> \<Longrightarrow>
       corres_rel' {(a, b). a = (b \<noteq> (0::word32))} (A >>= B) (A' >>= B')"
  apply atomize
  apply (monad_eq simp: corres_rel'_def)
  apply fastforce
  done


lemma "corres_rel {(a, b). a = (b \<noteq> 0)} dc (returnOk True) (returnOk (1 :: word32))"
  apply (monad_eq simp: corres_rel_def)
  done

definition "id_rel \<equiv> {(a, b). a = b}"


lemma "
  \<lbrakk> corres_rel {(a, b). a = (b \<noteq> 0)} id_rel A A';
    \<And>x x'. (x = (x' \<noteq> 0)) \<Longrightarrow> corres_rel {(a, b). a = (b \<noteq> (0 :: word32))} id_rel (B x) (B' x') \<rbrakk> \<Longrightarrow>
       corres_rel {(a, b). a = (b \<noteq> 0)} id_rel (A >>=E B) (A' >>=E B')"
  apply atomize
  apply (monad_eq simp: corres_rel_def id_rel_def split: sum.splits)
     apply (erule allE, erule impE, assumption)+
  apply safe
     apply (clarsimp split: sum.splits)
     apply (erule allE2, erule conjE, erule (1) impE)
     apply clarsimp
        apply (rule classical)
  oops

declare corresTA_def [simp]

lemma corresTA_L2_gets [word_abs]:
  "\<lbrakk> \<And>s. abstract_val (Q s) rx (C s) (C' s) \<rbrakk> \<Longrightarrow>
     corresTA rx ex (L2_seq (L2_guard Q) (\<lambda>_. (L2_gets C n))) (L2_gets C' n)"
  apply (monad_eq simp: L2_defs corresXF_def)
  done

lemma corresTA_L2_throw [word_abs]:
  "\<lbrakk> abstract_val Q ex C C' \<rbrakk> \<Longrightarrow>
     corresTA rx ex (L2_seq (L2_guard (\<lambda>_. Q)) (\<lambda>_. (L2_throw C n))) (L2_throw C' n)"
  apply (monad_eq simp: L2_defs corresXF_def)
  done

lemma corresTA_L2_skip [word_abs]:
  "corresTA rx ex L2_skip L2_skip"
  apply (monad_eq simp: L2_defs corresXF_def)
  done

(*** FIXME ***)
lemma corresTA_L2_seq [word_abs]:
  fixes L' :: "('s, 'e + 'c1) nondet_monad"
  fixes R' :: "'c1 \<Rightarrow> ('s, 'e + 'c2) nondet_monad"
  fixes L :: "('s, 'ea + 'a1) nondet_monad"
  fixes R :: "'a1 \<Rightarrow> ('s, 'ea + 'a2) nondet_monad"
  shows
  "\<lbrakk> (* valid_typ_abs_fn P rx1 crx1; *)
     corresTA rx1 ex L L';
     \<And>r. corresTA rx2 ex (R (rx1 r)) (R' r) \<rbrakk> \<Longrightarrow>
    corresTA rx2 ex
       (L2_seq L R)
       (L2_seq L' R')"
  apply atomize
  apply (clarsimp simp: L2_seq_def L2_guard_def)
  apply (erule corresXF_join [where P'="\<lambda>x y s. rx1 y = x"])
    apply (clarsimp simp: corresXF_def split: sum.splits)
    apply force
   apply clarsimp
   apply (rule hoareE_TrueI)
  apply simp
  done

lemma corresTA_L2_seq_unit [word_abs]:
  fixes L' :: "('s, 'e + unit) nondet_monad"
  fixes R' :: "unit \<Rightarrow> ('s, 'e + 'r) nondet_monad"
  fixes L :: "('s, 'ea + unit) nondet_monad"
  fixes R :: "('s, 'ea + 'ra) nondet_monad"
  shows
  "\<lbrakk> corresTA rx' ex L L';
     corresTA rx ex R (R' ()) \<rbrakk> \<Longrightarrow>
    corresTA rx ex
       (L2_seq L (\<lambda>_. R))
       (L2_seq L' R')"
  apply (rule corresTA_L2_seq)
   apply simp
  apply simp
  done


lemma corresTA_L2_catch  [word_abs]:
  fixes L' :: "('s, 'e1 + 'c) nondet_monad"
  fixes R' :: "'e1 \<Rightarrow> ('s, 'e2 + 'c) nondet_monad"
  fixes L :: "('s, 'e1a + 'ca) nondet_monad"
  fixes R :: "'e1a \<Rightarrow> ('s, 'e2a + 'ca) nondet_monad"
  shows
  "\<lbrakk> (* valid_typ_abs_fn P ex1 cex1; *)
     corresTA rx ex1 L L'; (*** FIXME ***)
     \<And>r. corresTA rx ex2 (R (ex1 r)) (R' r) \<rbrakk> \<Longrightarrow>
    corresTA rx ex2
       (L2_catch L R)
       (L2_catch L' R')"
  apply atomize
  apply (clarsimp simp: L2_seq_def L2_catch_def L2_guard_def)
  apply (erule corresXF_except [where P'="\<lambda>x y s. ex1 y = x"])
    apply (clarsimp simp: corresXF_def split: sum.splits cong: rev_conj_cong)
    apply force
   apply clarsimp
   apply (rule hoareE_TrueI)
  apply simp
  done

lemma corresTA_L2_while [word_abs]:
  assumes init_corres: "abstract_val Q rx i i'"
  and cond_corres: "\<And>r s. abstract_val (G (rx r) s) (\<lambda>x. x) (C (rx r) s) (C' r s)"
  and body_corres: "\<And>r. corresTA rx ex (B (rx r)) (B' r)"
  shows "corresTA rx ex
       (L2_seq (L2_guard (\<lambda>_. Q)) (\<lambda>_. L2_guarded_while (\<lambda>r s. G r s) C B i x))
       (L2_while C' B' i' x)"
proof -
  note body_corres' =
       corresXF_guarded_while_body [OF body_corres [unfolded corresTA_def]]

  have init_corres':
    "Q \<Longrightarrow> i = rx i'"
    using init_corres
    by simp

  show ?thesis
  apply (clarsimp simp: L2_defs guardE_def [symmetric] returnOk_liftE [symmetric])
  apply (rule corresXF_exec_abs_guard)
  apply (rule corresXF_guarded_while [where P="\<lambda>r s. (* P (rx r) \<and> *) G (rx r) s"])
        apply (cut_tac r=x in body_corres)
        apply (clarsimp simp: corresXF_def split: sum.splits)
        apply (insert cond_corres)[1]
        apply clarsimp
       apply clarsimp
       apply (rule hoareE_TrueI)
      apply (clarsimp simp: init_corres)
       apply (insert init_corres)[1]
       apply (clarsimp)
     apply (clarsimp simp: init_corres')
  done
qed

lemma corresTA_L2_guard [word_abs]:
  "\<lbrakk> \<And>s. abstract_val (Q s) (\<lambda>x. x) (G s) (G' s) \<rbrakk>
           \<Longrightarrow> corresTA rx ex (L2_guard (\<lambda>s. G s \<and> Q s)) (L2_guard G')"
  apply (monad_eq simp: L2_defs corresXF_def)
  done

lemma corresTA_L2_condition [word_abs]:
  "\<lbrakk> corresTA rx ex L L';
     corresTA rx ex R R';
     \<And>s. abstract_val (P s) (\<lambda>x. x) (C s) (C' s)  \<rbrakk>
   \<Longrightarrow> corresTA rx ex
             (L2_seq (L2_guard P) (\<lambda>_. L2_condition C L R)) (L2_condition C' L' R')"
  apply atomize
  apply (monad_eq simp: L2_defs corresXF_def Ball_def split: sum.splits)
  apply force
  done

lemma corresTA_L2_call [word_abs]:
  "corresTA a (\<lambda>x. x) (L2_seq (L2_call x y) (\<lambda>val. L2_gets (\<lambda>s. a val) [''ret''])) (L2_call x y)"
  apply (clarsimp simp: L2_defs L2_call_def corresXF_def)
  apply (monad_eq split: sum.splits)
  apply force
 done

lemma abstract_val_id [word_abs]:
  "abstract_val True (\<lambda>x. x) r r"
  by simp

lemma abstract_val_sint [word_abs]:
  "abstract_val True sint (sint n) (n :: sword32)"
  by clarsimp

lemma abstract_val_unat [word_abs]:
  "abstract_val True unat (unat r) (r :: word32)"
  apply (clarsimp simp: unat_of_nat_eq)
  done

lemma abstract_val_of_nat [word_abs]:
  "abstract_val (r \<le> UINT_MAX) unat r (of_nat r :: word32)"
  apply (clarsimp simp: unat_of_nat_eq UINT_MAX_def)
  done

lemma abstract_val_of_int [word_abs]:
  "abstract_val (INT_MIN \<le> r \<and> r \<le> INT_MAX) sint r (of_int r :: sword32)"
  by (clarsimp simp: sint_of_int_eq INT_MIN_def INT_MAX_def)

lemma valid_typ_abs_fn_sint [word_abs]:
  "valid_typ_abs_fn (\<lambda>v. INT_MIN \<le> v \<and> v \<le> INT_MAX) (sint :: sword32 \<Rightarrow> int) (of_int :: int \<Rightarrow> sword32)"
  apply (clarsimp simp: sint_of_int_eq INT_MIN_def INT_MAX_def)
  done

lemmas [word_abs] = word_abs_word_ops

end

