(*
 * Copyright (C) 2013, National ICT Australia Limited. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *  * The name of National ICT Australia Limited nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *)

(*
 * Construct necessary infrastructure to allow heap lifting to take
 * place.
 *
 * This includes things like:
 *
 *   * Creating new global records;
 *)

structure HeapLiftBase =
struct

type field_info = {
  name : string,
  field_type : typ,
  getter : term,
  setter : term
};

type struct_info = {
  name : string,
  struct_type : typ,
  field_info : field_info list
};

type heap_info =
{
  globals_type : typ,
  old_globals_type : typ,
  heap_getters : (string * typ) Typtab.table,
  heap_setters : (string * typ) Typtab.table,
  global_fields : (string * string * typ) list,
  global_field_getters : (term * term) Symtab.table,
  global_field_setters : (term * term) Symtab.table,
  lift_fn_name : string,
  lift_fn_thm : thm,
  lift_fn_full : term,
  dummy_state : term,
  lift_fn_simp_thms : thm list,
  structs : struct_info Symtab.table,
  struct_types : struct_info Typtab.table
}

(*
 * Attempt to generate a pleasant name for the record
 * field name for the given heap.
 *
 *   word32        => heap_w32
 *   word32 ptr    => heap_w32_ptr
 *   node_C ptr    => heap_node_ptr
 *)
fun heap_name_from_type (T : typ) =
let
  (* Convert a numerical type back into an integer. *)
  fun get_num_type x =
    case (dest_Type x) of
        (@{type_name "Numeral_Type.num0"}, _) => 0
      | (@{type_name "Numeral_Type.num1"}, _) => 1
      | (@{type_name "Numeral_Type.bit0"}, [T])
          => 2 * get_num_type T
      | (@{type_name "Numeral_Type.bit1"}, [T])
          => 2 * get_num_type T + 1

  fun get_name x =
    case (dest_Type x) of
         (@{type_name "Word.word"}, [x])
           => "w" ^ (PolyML.makestring (get_num_type x))
       | (@{type_name "ptr"}, [x])
           => (get_name x) ^ "'ptr"
       | (@{type_name "array"}, [x,y])
           => (get_name x) ^ "'array_" ^ (PolyML.makestring (get_num_type y))
       | (x, y) => (Long_Name.base_name x)
in
  "heap_" ^ (get_name T)
end

(*
 * Determine what types we will need for the split heap to support the given
 * term.
 *)
fun get_term_heap_types ctxt t =
let
  fun is_ptr_type T =
    case T of
        Type (@{type_name "ptr"}, [_]) => true
      | _ => false
in
  case t of
      (t as Abs _) => Utils.concrete_abs' ctxt t |> #1 |> get_term_heap_types ctxt
    | a $ b => get_term_heap_types ctxt a @ get_term_heap_types ctxt b
    | a => (body_type (fastype_of a) :: binder_types (fastype_of a))
           |> filter is_ptr_type
           |> map (dest_Type #> snd #> hd)
end

(* Determine what heap types the given program accesses. *)
fun get_program_heap_types prog_info fn_info lthy =
let
  (* Is the given heap type valid, or should we just filter it? *)
  fun is_valid_heap_type T =
    case T of
      (* "void *" pointers are not really heap types. *)
        @{typ "unit"} => false

      (* Signed words are converted to standard words on the heap. *)
      | Type (@{type_name "Word.word"},
          [Type (@{type_name "SignedWords.signed"}, _)]) => false

      | _ => true

  (* Process a function. *)
  fun process fn_name =
      FunctionInfo.get_function_def fn_info fn_name
      |> #definition
      |> prop_of
      |> Utils.rhs_of_eq
      |> get_term_heap_types lthy
      |> filter is_valid_heap_type
      |> Typset.make
in
  Typset.union_sets (AutoCorresUtil.map_all fn_info process)
end

(* Get fields from the globals type that should be copied from the
 * old globals type to the new globals type. *)
fun get_real_global_vars globalsT thy =
  (* Get all existing globals, filtering out "t_hrs_'". *)
   Record.get_recT_fields thy globalsT
   |> fst
   |> filter (fn (a, _) => Long_Name.base_name a <> "t_hrs_'")

(* Get the "t_hrs_'" accessor from the given globals record type. *)
fun get_globals_t_hrs globalsT thy =
   Record.get_recT_fields thy globalsT
   |> fst
   |> filter (fn (a, _) => Long_Name.base_name a = "t_hrs_'")
   |> hd


(*
 * Define a new heap record using the record package.
 *
 * "globalsT" is the type of the existing globals record, while "heapTs" is a
 * list of the different types the new heap will need to support.
 *)
fun gen_new_heap globalsT heapTs thy =
let
  val existing_fields = get_real_global_vars globalsT thy
  val new_heap_rec_name = "new_globals"

  (* Generate new fields. *)
  val copied_fields =
     map (fn (a, b) => (Binding.name (Long_Name.base_name a), b, NoSyn)) existing_fields
  val heap_fields =
     map (fn a => (Binding.name (heap_name_from_type a), mk_ptrT a --> mk_optionT a, NoSyn)) heapTs

  (* Define the record. *)
  val thy = Record.add_record ([], Binding.name new_heap_rec_name)
    NONE (copied_fields @ heap_fields) thy

  (* The record package doesn't tell us what we just definied, so we
   * attempt to recreate Fetch the type of the record. *)
  val full_rec_name = Sign.full_name thy (Binding.name new_heap_rec_name)
  val new_globalsT = Proof_Context.read_typ (Proof_Context.init_global thy) full_rec_name

  (* Create a list of the names of the new fields. *)
  val new_fields = Record.get_extT_fields thy new_globalsT
      |> fst |> take (length copied_fields)

  (* Hide the constants of existing fields. *)
  val thy = Isar_Cmd.hide_const true (map fst existing_fields) thy

  (* Generate a mapping from heap types to the getter/setter for that heap. *)
  val fields = Record.get_recT_fields thy new_globalsT |> fst
      |> drop (length copied_fields)
  val getters = map (fn (a,b) => (a, new_globalsT --> b)) fields
  val setters = map (fn (a,b) => (a ^ Record.updateN, (b --> b) --> (new_globalsT --> new_globalsT))) fields
  val type_heap_map = Typtab.make (heapTs ~~ (getters ~~ setters))
in
  ((new_globalsT, type_heap_map,
      map (fn ((a, T), (b, _)) => (a, b, T)) (existing_fields ~~ new_fields)), thy)
end

(*
 * Generate a heap abstraction function.
 *
 * That is, we generate a function that takes a "globals" variable that comes
 * from the C parser and spits out a "new_globals" variable which has a split
 * heap format.
 *
 * The "new_globals" type should be generated by "gen_new_heap".
 *)
fun gen_heap_abs_fn (prog_info : ProgramInfo.prog_info) old_globalsT new_globalsT heapTs lthy =
let
  val thy = Proof_Context.theory_of lthy
  val dummy_old_globals = Free ("g", old_globalsT)

  (* Fetch the fields of the globals record. *)
  val existing_fields = get_real_global_vars old_globalsT thy

  (* Fetch the "t_hrs_'" variable from the old globals, which
   * contains heap data. *)
  val t_hrs = get_globals_t_hrs old_globalsT thy

  (* Generate a term to construct the local variables record. *)
  val head_term = RecordUtils.get_record_constructor thy new_globalsT

  (*
   * We assume that the order of the variables in "new_globalsT"
   * match those in existing fields, followed by those in
   * heapTs.
   *
   * If this assumption is wrong, the following probably will
   * not type check (and, if it does, will be wrong).
   *
   * Our first step, we copy fields from the old globals type.
   *)
  val copy_term = fold (fn (name, t) => fn rest =>
    (rest $ (Const (name, old_globalsT --> t) $ dummy_old_globals)))
      existing_fields head_term

  (* Next, we generate lifted heaps from the old globals type. *)
  val lift_term = fold (fn t => fn rest =>
    (rest $ (Abs ("x", mk_ptrT t,
      Const (@{const_name "simple_lift"}, dummyT)
        $ (Const (fst t_hrs, dummyT) $ dummy_old_globals) $ Bound 0))))
      (Typset.dest heapTs) copy_term
     |> (fn t => t $ @{term "()"})
     |> Syntax.check_term lthy

  (* Generate a body of a definition, and define it. *)
  val (Const (lift_name, _), def_thm, lthy) =
    Utils.define_const_args "lift_global_heap" false
       lift_term [("g", old_globalsT)] lthy

  (* Generate some simp rules to make life easier. *)
  val (simp_thms, lthy) = RecordUtils.generate_ext_simps "new_globals_ext_simps" def_thm lthy
in
  (lift_name, def_thm, simp_thms, lthy)
end

(*
 * Fetch information about structures in the program, such as
 * the fields of each structure and their types.
 *
 * We return a tuple containing (
 *     <mapping from struct names to field information>,
 *     <mapping from struct typs to field information>
 * ).
 *)
fun get_prog_struct_info thy prog_info =
let
  (* Fetch the namespace we are working in. *)
  val namespace =
    Long_Name.explode (dest_Type (#globals_type prog_info) |> fst)
    |> hd |> Long_Name.explode |> hd

  (* Fetch information about structures defined in the program. *)
  val struct_data = ProgramAnalysis.get_senv (#csenv prog_info)
        @ ProgramAnalysis.get_globals_rcd (#csenv prog_info)

  (* Given the name of a struct ("struct moo" would be "moo_C"), return
   * the type of that structure. *)
  fun get_struct_type name =
    CalculateState.ctype_to_typ (thy, CTypeDatatype.StructTy name)

  (* Generate information relating to a field of a single struct. *)
  fun get_field_info struct_name struct_type (field_name, field_type) = let
    val hol_field_type = CalculateState.ctype_to_typ (thy, field_type)
  in
    {
      name = field_name,
      field_type = hol_field_type,
      getter =
        Const (Long_Name.implode [namespace, struct_name, field_name],
            struct_type --> hol_field_type),
      setter =
        Const (Long_Name.implode [namespace, struct_name, field_name ^ "_update"],
            (hol_field_type --> hol_field_type) --> struct_type --> struct_type)
    }
  end

  (* Generate info relating to a single struct. *)
  fun get_struct_info (struct_name, struct_fields) =
  let
    val struct_type = get_struct_type struct_name
  in
    (struct_name, {
      name = struct_name,
      struct_type = struct_type,
      field_info = map (get_field_info struct_name struct_type) struct_fields
    })
  end

  val struct_info = map get_struct_info struct_data
in
  (Symtab.make struct_info,
      Typtab.make (map (fn (a,b) => (get_struct_type a, b)) struct_info))
end

fun mk_heap_info
  (thy : theory)
  (prog_info : ProgramInfo.prog_info)
  (new_globalsT : typ)
  (heap_getters_setters : ((string * typ) * (string * typ)) Typtab.table)
  (lift_fn : string * thm)
  (simp_thms : thm list)
  (global_fields : (string * string * typ) list) =
let
  val old_globalsT = #globals_type prog_info
  val (structs, struct_types) = get_prog_struct_info thy prog_info
  fun get_getter_const name globalsT destT
      = Const (name, globalsT --> destT)
  fun get_setter_const name globalsT destT
      = Const (name ^ Record.updateN, (destT --> destT) --> globalsT --> globalsT)
in
  {
    (* Type of the new globals state. *)
    globals_type = new_globalsT,
    old_globals_type = old_globalsT,

    (* Heap getters / setters. *)
    heap_getters = Typtab.map (K fst) heap_getters_setters,
    heap_setters = Typtab.map (K snd) heap_getters_setters,

    (* List of mappings between the old and new global fields. *)
    global_fields = global_fields,
    global_field_getters =
        map (fn (old_name, new_name, T) =>
          (old_name, (get_getter_const old_name old_globalsT T,
            get_getter_const new_name new_globalsT T)))
          global_fields
        |> Symtab.make,
    global_field_setters=
        map (fn (old_name, new_name, T) =>
          (old_name, (get_setter_const old_name old_globalsT T,
            get_setter_const new_name new_globalsT T)))
          global_fields
        |> Symtab.make,

    (* Function to lift the old globals type into the new globals type. *)
    lift_fn_name = fst lift_fn,
    lift_fn_thm = snd lift_fn,

    (* Function to lift old globals into new globals. *)
    lift_fn_full = Const (fst lift_fn, old_globalsT --> new_globalsT),

    (* Dummy state variable, used as a placeholder during translation. *)
    dummy_state = Free ("_dummy_state", new_globalsT),

    (* Simplification theorems for the lifting function. *)
    lift_fn_simp_thms = simp_thms,

    (* Structure information. *)
    structs = structs,
    struct_types = struct_types
  } : heap_info
end

fun setup prog_info fn_info lthy =
let
  val old_globalsT = #globals_type prog_info

  (* Generate a new globals structure. *)
  val heapTs = get_program_heap_types prog_info fn_info lthy
  val ((new_globalsT, heap_getters_setters, global_fields), lthy)
      = Local_Theory.raw_theory_result (
          gen_new_heap old_globalsT (Typset.dest heapTs)) lthy

  (*
   * HACK: Exit and enter the context again, so the simpset created by the new
   * record gets imported in. No idea what the "correct" way of doing this is.
   *)
  val lthy = Named_Target.reinit lthy lthy

  (* Generate a function mapping old globals to the new globals. *)
  val (lift_name, lift_def, simp_thms, lthy) = gen_heap_abs_fn
      prog_info old_globalsT new_globalsT heapTs lthy

  (* Generate data structure encoding all relevant information. *)
  val heap_info = mk_heap_info (Proof_Context.theory_of lthy) prog_info new_globalsT
        heap_getters_setters (lift_name, lift_def) simp_thms
        global_fields
in
  (heap_info, lthy)
end


end
