(*
 * Copyright (C) 2013, National ICT Australia Limited. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *  * The name of National ICT Australia Limited nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *)
(*
 * Optimise L2 fragments of code by using facts learnt earlier in the fragments
 * to simplify code afterwards.
 *)

structure L2Opt =
struct

(*
 * Solve a goal of the form:
 *
 *   simp_expr P A ?X
 *
 * This is done by simplifing "A" while assuming "P", and unifying the result
 * (usually instantiating "X") in the process.
 *)
val simp_expr_thm =
    @{lemma "(simp_expr P G G == simp_expr P G G') ==> simp_expr P G G'" by (clarsimp simp: simp_expr_def)}
fun solve_simp_expr_tac ctxt =
  Subgoal.FOCUS_PARAMS (fn {context = ctxt, schematics = (_, term_schem), ...} =>
  (fn thm =>
    case Drule.cprems_of thm of
      [] => (no_tac thm)
    | (goal::_) =>
      (case term_of goal of
        (_ $ (Const (@{const_name "simp_expr"}, _) $ P $ L $ _)) =>
        let
          val goal = @{mk_term "simp_expr ?P ?L ?L" (P, L)} (P, L)
              |> cterm_of (Proof_Context.theory_of ctxt)
          val ss = simpset_of ctxt
                   |> Simplifier.add_cong @{thm split_cong}
                   |> Simplifier.add_cong @{thm if_cong}
                   |> Simplifier.add_cong @{thm conj_cong}
          val simplified = Simplifier.asm_full_rewrite ss goal

          (* Ensure that all schematics have been resolved. *)
          val schematic_remains = Term.exists_subterm Term.is_Var (prop_of simplified)
        in
          if schematic_remains then
            (rtac @{thm simp_expr_triv} 1) thm
          else
            ((rtac simp_expr_thm 1) THEN (rtac simplified 1)) thm
        end
      | _ => no_tac thm)
    )) ctxt

(*
 * Solve a goal of the forms:
 *
 *   simp_expr P A B
 *
 * where both "A" and "B" are constants (i.e., not schematics).
 *)
fun solve_simp_expr_const_tac ctxt thm =
  if (Term.exists_subterm Term.is_Var (term_of (cprem_of thm 1))) then
    no_tac thm
  else
    SOLVES (
      (rtac @{thm simp_expr_solve_constant} 1)
      THEN (Clasimp.clarsimp_tac
        (map_simpset (fn ss => ss addsimps [@{thm split_def}]) ctxt) 1)) thm

(*
 * Recursively simplify a monadic expression, using information gleaned from
 * earlier in the program to simplify parts of the program further down.
 *)
fun monad_equiv ctxt ct =
let
  (* Mark context as being "invisible" to reduce warnings being printed. *)
  val ctxt = Context_Position.set_visible false ctxt
  val thy = Proof_Context.theory_of ctxt

  (* Generate our top-level "monad_equiv" goal. *)
  val goal = @{mk_term "?L == ?R" (L)} (term_of ct)
    |> cterm_of thy
    |> Goal.init
    |> Utils.apply_tac "Creating object-level equality." (rtac @{thm eq_reflection} 1)
    |> Utils.apply_tac "Creating 'monad_equiv' goal." (rtac @{thm monad_equiv_eq} 1)

  (* Print a diagnostic if this branch fails. *)
  val num_failures = ref 0
  fun print_failure_tac t =
    if (false andalso !num_failures < 5) then
      (num_failures := !num_failures + 1; (print_tac "Branch failed" THEN no_tac) t)
    else
      (no_tac t)

  (* Fetch theorms used in the simplification process. *)
  val thms = L2FlowThms.get ctxt

  (* Tactic to blindly apply simplification rules. *)
  fun solve_goal_tac _ =
    DETERM (SOLVES
      ((solve_simp_expr_const_tac ctxt)
      ORELSE
      ((solve_simp_expr_tac ctxt 1)
      ORELSE
      ((resolve_tac thms THEN_ALL_NEW solve_goal_tac) 1
      ORELSE
      ((print_failure_tac))))))

  (* Apply the rules. *)
  val thm =
    Utils.apply_tac "Simplifying L2" (solve_goal_tac 1) goal
    |> Goal.finish ctxt
in
  thm
end

(*
 * A simproc implementing the "L2_gets_bind" rule. The rule, unfortauntely, has
 * the ability to cause exponential growth in the spec size in some cases;
 * thus, we can only selectively apply it in cases where this doesn't happen.
 *
 * In particular, we propagate a "gets" into its usage if it is used only once.
 *)
val l2_gets_bind_thm = mk_meta_eq @{thm L2_gets_bind}
fun l2_gets_bind_simproc' thy simpset term =
let
  fun is_simple (_ $ Abs (_, _, Bound _)) = true
    | is_simple (_ $ Abs (_, _, Free _)) = true
    | is_simple (_ $ Abs (_, _, Const _)) = true
    | is_simple x = false
in
  case term of
    (Const (@{const_name "L2_seq"}, _) $ lhs $ Abs (n, _, rhs)) =>
      let
        fun count_var_usage (a $ b) = count_var_usage a + count_var_usage b
          | count_var_usage (Abs (_, _, x)) = count_var_usage x
          | count_var_usage (Free ("_dummy", dummyT)) = 1
          | count_var_usage _ = 0
        val count = count_var_usage (subst_bounds ([Free ("_dummy", dummyT)], rhs))
      in
        if count <= 1 orelse is_simple lhs then
          SOME l2_gets_bind_thm
        else
          NONE
      end
    | _ => NONE
end
val l2_gets_bind_simproc =
  Raw_Simplifier.mk_simproc "L2_gets_bind_simproc"
      [@{cpat "L2_seq (L2_gets (%_. ?A) ?n) ?B"}] l2_gets_bind_simproc'

(* Simproc to clean up guards. *)
fun l2_guard_simproc' full_ss thy _ term =
let
  val simp_thm = Simplifier.asm_full_rewrite
      (Simplifier.add_cong @{thm rev_conj_cong} full_ss) (cterm_of thy term)
  val [lhs, rhs] = Thm.prop_of simp_thm |> Term.strip_comb |> snd
in
  if Term_Ord.fast_term_ord (lhs, rhs) = EQUAL then
    NONE
  else
    SOME simp_thm
end
fun l2_guard_simproc full_ss =
  Raw_Simplifier.mk_simproc "L2_guard_simproc"
    [@{cpat "L2_guard ?G"}] (l2_guard_simproc' full_ss)

(*
 * Carry out flow-sensitive optimisations on the given 'thm'.
 *
 * 'n' is the argument number to cleanup, counting from 1. So for example, if
 * our input theorem was "corres P A B", a "n" of 3 would simplify "A".
 *
 * If "fast_mode" is true, we don't do flow-sensitive optimisations (which tend
 * to be time-consuming).
 *)
fun cleanup_thm ctxt thm fast_mode n =
let
  (* Don't print out warning messages. *)
  val ctxt = Context_Position.set_visible false ctxt

  (* Setup basic simplifier. *)
  val solver = Raw_Simplifier.mk_solver "simp solver"
      (K (asm_full_simp_tac (Simplifier.simpset_of ctxt)))
  val ss = Utils.hidden_HOL_basic_ss
      addsimps (L2PeepholeThms.get ctxt)
      addsimprocs [l2_gets_bind_simproc,
          l2_guard_simproc (Simplifier.simpset_of ctxt)]
      setSolver solver
  val late_ss = ss addsimps (L2PeepholeLateThms.get ctxt)
  val simp = (Simplifier.rewrite ss) then_conv (Simplifier.rewrite late_ss)
  fun l2conv conv = (Conv.arg_conv (Utils.nth_arg_conv n conv))

  (* Apply peephole optimisations to the theorem. *)
  val new_thm = Conv.fconv_rule (l2conv simp) thm
      |> Drule.eta_contraction_rule

  (* Apply flow-sensitve optimisations, and the re-apply simple simplifications. *)
  val new_thm =
    if not fast_mode then
    let
      val ct = new_thm |> cprop_of |> Utils.ctail_of |> (fn x => Utils.cterm_nth_arg x (n - 1))
      val result = monad_equiv ctxt ct
    in
      Conv.fconv_rule (l2conv (Conv.rewr_conv result)) new_thm
      |> Conv.fconv_rule (l2conv simp)
    end
    else
      new_thm

  (* Beta/Eta normalise. *)
  val new_thm = Conv.fconv_rule (l2conv Thm.eta_conversion) new_thm
  val new_thm = Conv.fconv_rule (l2conv (Thm.beta_conversion true)) new_thm
in
  new_thm
end

end
