(*
 * Copyright (C) 2013, National ICT Australia Limited. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *  * The name of National ICT Australia Limited nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *)

(*
 * Code to manage converting between L2_monad and other monad types.
 *
 * TypeStrengthen provides a higher level interface for converting entire programs.
 *)

structure Monad_Convert = struct

(* Utilities. *)
fun intersperse _ [] = []
  | intersperse _ [x] = [x]
  | intersperse a (x::xs) = x :: a :: intersperse a xs

fun theE NONE exc = raise exc
  | theE (SOME x) _ = x

fun oneE [] exc = raise exc
  | oneE (x::_) _ = x



(* From Find_Theorems *)
fun apply_dummies tm =
  let
    val (xs, _) = Term.strip_abs tm;
    val tm' = Term.betapplys (tm, map (Term.dummy_pattern o #2) xs);
  in #1 (Term.replace_dummy_patterns tm' 1) end;

fun parse_pattern ctxt nm =
  let
    val consts = Proof_Context.consts_of ctxt;
    val nm' =
      (case Syntax.parse_term ctxt nm of
        Const (c, _) => c
      | _ => Consts.intern consts nm);
  in
    (case try (Consts.the_abbreviation consts) nm' of
      SOME (_, rhs) => apply_dummies (Proof_Context.expand_abbrevs ctxt rhs)
    | NONE => Proof_Context.read_term_pattern ctxt nm)
  end;

(* Breadth-first term search *)
fun term_search_bf cont pred prune = let
  fun search (term, queue) =
    if pred term then cont term (fn () => walk queue) else
    if prune term then walk queue else
    case term of
        Abs (_, _, t) => walk (Queue.enqueue t queue)
      | f $ x => walk (Queue.enqueue x (Queue.enqueue f queue))
      | _ => walk queue
  and walk queue = if Queue.is_empty queue then () else search (Queue.dequeue queue)
in
  (fn term => search (term, Queue.empty))
end

fun term_search_bf_first pred prune term = let
  val r = Unsynchronized.ref NONE
  val _ = term_search_bf (fn term => K (r := SOME term)) pred prune term
in !r end

fun grep_term ctxt pattern =
let
  val thy = Proof_Context.theory_of ctxt
in
  term_search_bf_first
      (fn term => Pattern.matches thy (pattern, term))
      (fn term => not (Pattern.matches_subterm thy (pattern, term)))
end

(* Check whether the term is in L2_monad notation. *)
val term_is_L2 = Monad_Types.check_lifting_head
    [@{term "L2_unknown"}, @{term "L2_seq"}, @{term "L2_modify"},
     @{term "L2_gets"}, @{term "L2_condition"}, @{term "L2_catch"}, @{term "L2_while"},
     @{term "L2_throw"}, @{term "L2_spec"}, @{term "L2_guard"}, @{term "L2_fail"},
     @{term "L2_recguard"}, @{term "L2_call"}]

(*
 * Perform monad conversion on a term, taking into account any extra
 * simplifying facts. Only a successful conversion is returned.
 *
 * For this conversion to be useful on recursive programs, it needs
 * to be given a fact representing the inductive assumption.
 *)
fun monad_rewrite (lthy : Proof.context) (mt : Monad_Types.monad_type)
                  (more_facts : thm list) (forward : bool)
                  (term : term) : thm option =
let
  val rules = if forward then #lift_rules mt else #unlift_rules mt
  val rules' = dest_ss rules |> #simps |> map #2
  val simps = HOL_ss addsimps rules' addsimps more_facts
  val cterm = cterm_of (Proof_Context.theory_of lthy) term
  (* Just apply the simplifier and hope that it works. *)
  val thm = Simplifier.rewrite simps cterm
  val rhs = Utils.rhs_of (term_of_thm thm)
  val good_rewrite = if forward then #valid_term mt else term_is_L2
in
  if good_rewrite lthy rhs
    then SOME thm else NONE
end

(*
 * Apply polish to a theorem of the form:
 *
 *   <LHS> == <lift> $ <some term to polish>
 *
 * Return the new theorem.
 *)
fun polish ctxt (mt : Monad_Types.monad_type) thm =
let
  (* Apply any polishing rules. *)
  val solver = Raw_Simplifier.mk_solver "simp solver"
      (K (asm_full_simp_tac (Simplifier.simpset_of ctxt)))
  val ss = Utils.merge_ss_ctxt (HOL_basic_ss addsimps (PolishSimps.get ctxt),
                                #polish_rules mt)
             setSolver solver
  val thm_p = Conv.fconv_rule (Conv.arg_conv (Utils.rhs_conv (Simplifier.rewrite ss))) thm
in
  thm_p
end


(*
 * monad_convert tactic.
 *)
fun monad_convert_tac (ctxt : Proof.context) (monad_name : string)
                      (pattern_str : string) (n : int) : tactic =
  fn state => let
    val all_rules = Monad_Types.TSRules.get (Context.Proof ctxt)

    (* Figure out which monad to lift into. *)
    val target_rule = theE (Symtab.lookup all_rules monad_name)
                           (ERROR ("monad_convert: could not find monad type " ^ quote monad_name))

    (* Search the subgoal for the supplied pattern. *)
    val pattern = parse_pattern ctxt pattern_str
    val subgoal = Logic.get_goal (term_of_thm state) n
    val m_term = theE (grep_term ctxt pattern subgoal)
                      (ERROR "monad_convert: failed to match pattern")

    (* Find a lifting rule whose output matches m_term.
     * This saves us from having to try every unlift rule. *)
    val orig_lift_rule = oneE (filter (fn mt => #valid_term mt ctxt m_term)
                                      (all_rules |> Symtab.dest |> map snd))
                              (TERM ("monad_convert: could not determine monad type", [m_term]))

    (* Unlift back to L2_monad. *)
    val unlift_thm = theE (monad_rewrite ctxt orig_lift_rule [] false m_term)
                          (TERM ("monad_convert: could not unlift term (rule: " ^
                                   #name orig_lift_rule ^ ")", [m_term]))
    val unlift_term = Utils.rhs_of (term_of_thm unlift_thm)

    (* Lift to target monad. *)
    val relift_thm = theE (monad_rewrite ctxt target_rule [] true unlift_term)
                          (TERM ("monad_convert: could not lift to " ^ #name target_rule,
                                 [m_term, unlift_term]))

    (* Polish result. *)
    val relift_thm' = polish ctxt target_rule relift_thm

    val translate_thm = Thm.transitive unlift_thm relift_thm'

    (* HACK: turn type variables into schematics *)
    val translate_thm' = Goal.prove ctxt [] [] (term_of_thm translate_thm) (K (rtac translate_thm 1))
  in
    EqSubst.eqsubst_tac ctxt [0] [translate_thm'] n state
  end

val _ = Context.>> (Context.map_theory
          (Method.setup (Binding.name "monad_convert")
            (* Based on subgoal_tac parser *)
            (Args.goal_spec -- Scan.lift (Parse.name -- Args.name_source) >>
            (fn (quant, (monad_name, term_str)) => fn ctxt =>
                SIMPLE_METHOD'' quant (monad_convert_tac ctxt monad_name term_str)))
           "autocorres monad conversion"))

end
