(*
 * Copyright (C) 2013, National ICT Australia Limited. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *  * The name of National ICT Australia Limited nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *)
(*
A story of AutoCorres and recursion.

To do:
-- prove SIMPL total correctness
-- remove (* slow *) steps
-- remove word32 metis (or wait for word lifting...)
-- do fib'_spec with Hoare logic and wp
-- Isar style proof?
*)

theory Fib
imports
  "../../AutoCorres"
  "../../../lib/OptionMonadWP"
begin

(*
The venerable Fibonacci function.
*)
fun fibo :: "nat \<Rightarrow> nat" where
  "fibo 0 = 0" |
  "fibo (Suc 0) = Suc 0" |
  "fibo (Suc (Suc n)) = fibo n + fibo (Suc n)"
declare fibo.simps [simp del]

lemma fibo_alt_def: "fibo n = (if n = 0 then 0 else if n = 1 then 1 else fibo (n - 1) + fibo (n - 2))"
  apply (induct n rule: less_induct)
  apply (rename_tac n, case_tac n, simp add: fibo.simps)
  apply (rename_tac n1, case_tac n1, simp add: fibo.simps)
  apply (simp add: fibo.simps)
  done

(*
We implement it twice in C, and compare correctness proofs on the
SIMPL and AutoCorres embeddings.
*)


(*
Now, C arithmetic is done using `unsigned', which is translated as `word32'.
So, it is much easier to prove that the C code implements this function,
which is not quite the same function as fibo.
*)
function fibo32 :: "word32 \<Rightarrow> word32" where
  "fibo32 n = (if n = 0 then 0 else if n = 1 then 1 else fibo32 (n - 1) + fibo32 (n - 2))"
apply auto
done
termination fibo32
  by (relation "measure (\<lambda>x. unat x)", (simp|unat_arith)+)
declare fibo32.simps [simp del]

(*
But we really want to say that the C code does implement fibo
(at least up till 2971215073 < 2^32)...
*)
lemma fibo_superlinear: "(6 + n) < fibo (6 + n)"
  apply (induct n rule: less_induct)
  apply (subst fibo_alt_def)
  apply simp
  apply (rename_tac n, case_tac n, simp, eval)
  apply (rename_tac n1, case_tac n1, simp, eval)
  apply (rename_tac n2)
  apply (rule_tac t="4 + n" and s="6 + n2" in subst)
   apply simp
  apply (rule_tac t="5 + n" and s="6 + n1" in subst)
   apply simp
  apply (subgoal_tac "6 + n < (6 + n1) + (6 + n2)")
   apply (subgoal_tac "n1 < n" "n2 < n")
     apply (subgoal_tac "\<And>(l::nat) x y gx gy. \<lbrakk> l < x + y; x < gx; y < gy \<rbrakk> \<Longrightarrow> l < gx + gy")
      apply blast
     apply arith
    apply simp_all
  done
lemma fibo_superlinear': "n \<ge> 6 \<Longrightarrow> n < fibo n"
  by (metis le_iff_add fibo_superlinear)
lemma unat_word32_plus: "unat x + unat y < 2^32 \<Longrightarrow> unat x + unat y = unat (x + y :: word32)"
  by (metis unat_of_nat32 word_arith_nat_add word_bits_conv)

(* ... so we should say that too. *)
lemma fibo32_is_fibo: "fibo n < 2^32 \<Longrightarrow> fibo n = unat (fibo32 (of_nat n))"
  apply (induct n rule: less_induct)
  apply (subst fibo32.simps)
  apply (subst fibo_alt_def)
  apply (rename_tac n, case_tac "n = 0", simp)
  apply (case_tac "n = 1", simp)
  apply (subgoal_tac "n < 2^32"
                     "fibo (n - 1) + fibo (n - 2) < 2^32"
                     "of_nat n \<noteq> (0 :: word32)"
                     "of_nat n \<noteq> (1 :: word32)"
                     "fibo (n - 1) < 2^32"
                     "fibo (n - 2) < 2^32"
                     "fibo (n - 1) = unat (fibo32 (of_nat (n - 1)))"
                     "fibo (n - 2) = unat (fibo32 (of_nat (n - 2)))")
          using unat_word32_plus
          apply simp
         apply (metis diff_less gr0I zero_less_numeral)
        apply (metis diff_less gr0I zero_less_one)
       apply simp
      apply simp
     apply (metis unat_1 unat_of_nat32 word_bits_conv)
    apply (metis unat_0 unat_of_nat32 word_bits_conv)
   apply (metis fibo_alt_def)
  apply (case_tac "n < 6")
   apply simp
  apply (subgoal_tac "n < fibo n")
   apply simp
  apply (simp add: fibo_superlinear')
done

(* A helper for the SIMPL proofs later. *)
lemma fibo32_rec: "\<lbrakk> a < a + 2; b = a + 1; c = a + 2 \<rbrakk> \<Longrightarrow> fibo32 a + fibo32 b = fibo32 c"
  apply (subst(3) fibo32.simps)
  apply simp
  apply safe
     apply unat_arith
    apply (metis comm_semiring_1_class.normalizing_semiring_rules(24) less_is_non_zero_p1)
   apply (metis word_not_simps(1))
  apply (simp add: field_simps)
  done



(* First we invoke CParser to translate the C code to SIMPL. *)
install_C_file "fib.c"


context fib begin
(* fib_linear\<^bsub>C\<^esub> is the linear-time implementation. *)
thm fib_linear_body_def
(* fib\<^bsub>C\<^esub> is the pretty (inefficient) recursive implementation. *)
thm fib_body_def
(*
Actually, neither function looks that pretty in SIMPL.
But we should be able to prove that they implement fibo32.
*)

(* First, the linear version. *)
lemma fib_linear_simpl_spec:
"\<Gamma> \<turnstile> {s. s = t}
     \<acute>ret__unsigned :== CALL fib_linear(\<acute>n)
     \<lbrace> (\<acute>ret__unsigned = fibo32 \<^bsup>t\<^esup>n) \<rbrace>"
  (* We have not annotated the code yet, so we cannot apply vcg usefully. *)
  (* First we expand the function call and defer the overall precondition. *)
  apply vcg_step
   defer
   (* Now we annotate the loop with the correct invariant and termination measure. *)
   apply (subst whileAnno_def)
   apply (subst whileAnno_def [symmetric,
     where I=" \<lbrace> \<acute>a = fibo32 (\<^bsup>t\<^esup>n - \<acute>n) \<and> \<acute>n \<le> \<^bsup>t\<^esup>n \<and> (\<acute>n \<noteq> 0 \<longrightarrow> (\<acute>b = fibo32 (\<^bsup>t\<^esup>n + 1 - \<acute>n))) \<rbrace>"
     and V="measure (\<lambda>s. unat (n_' s))"])
   apply vcg
  (* It is mostly word arithmetic from here on. *)
    apply (simp add: scast_def field_simps)
    apply clarsimp
    apply (case_tac "n = 0")
     apply clarsimp
    apply (case_tac "n = 1")
     apply (rename_tac n1, subgoal_tac "n1 = 1")
      apply simp
     apply unat_arith
    apply (rename_tac n1, case_tac "n1 = 1")
     apply simp
    apply clarsimp
    apply safe
       apply (metis linear word_must_wrap)
      apply (rule fibo32_rec)
        apply unat_arith (* slow *)
       apply (simp add: field_simps)+
   apply (simp add: fibo32.simps)
  apply (simp add: scast_def fibo32.simps)
done

(* And the recursive version. *)
thm fib_body_def

lemma fib_simpl_spec: "\<forall>n. \<Gamma>,\<Theta>\<turnstile>\<^sub>t\<lbrace>\<acute>n=n\<rbrace>  PROC fib(\<acute>n,\<acute>ret__unsigned) \<lbrace>\<acute>ret__unsigned = fibo32 n\<rbrace>"
  apply (hoare_rule HoareTotal.ProcRec1[where r = "measure (\<lambda>(s, d). unat \<^bsup>s\<^esup>n)"])
  apply (unfold creturn_def | vcg_step)+
  apply (subst fibo32.simps, simp)
  apply (subst fibo32.simps, simp add: scast_id)
  apply (subst fibo32.simps[where n = n])
  apply unat_arith
  done



(*
We need to temporarily leave the local context to run autocorres.
Normally, we would run autocorres immediately after install_C_file.
*)
end (* context fib *)

(*
AutoCorres will perform a series of translations,
lifting the SIMPL code into Haskell-like monadic code.

NB: AutoCorres would also lift fib_linear\<^bsub>C\<^esub> to the option monad.
But for demonstration, we force it to a different monad here.
*)
autocorres [ ts_force nondet = fib_linear ] "fib.c"
context fib begin
(* fib_linear\<^bsub>C\<^esub> is lifted into a state monad; *)
thm fib_linear'_def
(* fib\<^bsub>C\<^esub> is lifted into an option monad. *)
thm fib'.simps
(* Note that fib_linear' and fib' are functions in HOL. *)


(* Let us prove that fib_linear' implements fibo32. *)
lemma fib_linear'_correct: "\<lbrace> \<lambda>s. P s \<rbrace> fib_linear' n \<lbrace> \<lambda>r s. r = fibo32 n \<and> P s \<rbrace>"
  apply (unfold fib_linear'_def)
  (* The loop invariant, as before. *)
  apply (subst whileLoop_add_inv
    [where I="\<lambda>(a', b', n') s. a' = fibo32 (n - n') \<and> n' \<le> n \<and> (n' \<noteq> 0 \<longrightarrow> (b' = fibo32 (n - n' + 1))) \<and> P s"])
  apply wp
    apply safe
        (* Clean up after the word arithmetic. *)
        apply (simp add: field_simps)
       apply (metis linear word_must_wrap)
      apply (subst fibo32.simps[where n = "n - (b - 1) + 1"])
      apply (subgoal_tac "n - (b - 1) + 1 \<noteq> 0")
       apply (subgoal_tac "n - (b - 1) + 1 \<noteq> 1")
        apply simp
        apply (metis (hide_lams, no_types) comm_semiring_1_class.normalizing_semiring_rules(24) is_num_normalize(8))
       apply (metis (hide_lams, no_types) add_diff_cancel2 comm_semiring_1_class.normalizing_semiring_rules(24) eq_iff_diff_eq_0 less_1_simp not_less_iff_gr_or_eq)
      apply (metis le_less less_1_simp word_sub_plus_one_nonzero)
     apply (simp_all add: fibo32.simps)
  done

(* We haven't actually proved that fib_linear' terminates. We can do that separately. *)
lemma fib_linear'_term: "no_fail (\<lambda>s. P s) (fib_linear' n)"
  (* We would like to use the wp tactic, so we will translate our goal back into
     Hoare syntax. *)
  apply (subst no_fail_is_validNF_True)
  apply (unfold fib_linear'_def)
  apply (subst whileLoop_add_inv [where M = "\<lambda>(_, _, n') s. n'"])
  apply wp
    apply simp
   apply (simp add: hoare_return_simp split: prod.splits)
   apply unat_arith
  apply simp
  done

(* And we are done. *)
lemma fib_linear'_spec: "\<lbrace> \<lambda>s. P s \<rbrace> fib_linear' n \<lbrace> \<lambda>r s. r = fibo32 n \<and> P s\<rbrace>!"
  by (rule validNF, rule fib_linear'_correct, rule fib_linear'_term)


(*
Now for the recursive function fib\<^bsub>C\<^esub>.
The lifted function fib' is a recursive function in HOL, but AutoCorres
cannot automatically prove termination, which is required by Isabelle/HOL.
Instead, fib' gets an extra parameter called `measure', which limits the
recursion depth. With this additional parameter, it is obvious that
fib' terminates, so AutoCorres can define fib' as a HOL function.
*)
thm fib'.simps (* Just a reminder *)
(*
This is also why fib\<^bsub>C\<^esub> is lifted to the option monad. If the measure parameter
is too small, fib' will return with failure instead of the result that would be
returned by fib\<^bsub>C\<^esub>.
*)

(*
AutoCorres also generates a \<^bitalic>measure function\<^eitalic> for fib'. This function is
used to generate the measure parameter when fib' is called. For example,

  void call_fib(void) { fib(42); }

is translated to:
*)
thm call_fib'_def
(*
The measure function receives all of the information needed, in principle,
to determine an upper bound on the call depth (if there is one).
Namely, the function arguments and the global program state.
*)
term measure_of'_fib
(*
measure_of'_fib does not actually have a definition, yet.
To give it a sensible definition, we need to work out the maximum
recursion depth, and AutoCorres does not know how to do that.
We will manually define the function later.
*)


(* But first, let us prove the correctness of fib'. *)
declare fib'.simps [simp del]

(*
Again, we prove partial correctness separately, assuming that fib'
does not return None.
*)

lemma fib'_correct_rec_helper:
  assumes ind: "\<And>y P m. ovalid (\<lambda>_. y < x \<and> P (fibo32 y)) (fib' m y) (\<lambda>r _. P r)"
  shows "ovalid (\<lambda>_. P (fibo32 x)) (fib' m x) (\<lambda>r _. P r)"
  apply (subst fib'.simps)
  apply (wp ind)
  apply simp
  apply safe
      apply (simp add: fibo32.simps | unat_arith)+
  done

lemma fib'_correct: "ovalid P (fib' m n) (\<lambda>r s. P s \<and> r = fibo32 n)"
  apply (subgoal_tac "\<And>P. ovalid (\<lambda>_. P (fibo32 n)) (fib' m n) (\<lambda>r _. P r)")
   apply (unfold ovalid_def, fast)[1]
  apply (induct n arbitrary: P m rule: less_induct)
  apply (subgoal_tac "\<And>y P m. ovalid (\<lambda>_. y < x \<and> P (fibo32 y)) (fib' m y) (\<lambda>r _. P r)")
   apply (rule fib'_correct_rec_helper)
   apply (unfold ovalid_def, fast)[1]
  apply (unfold ovalid_def)
  apply blast
  done

lemma fib'_term_rec_helper:
  assumes ind: "\<And>y P measure'. ovalidNF (\<lambda>_. y < x \<and> unat y < measure' \<and> P) (fib' measure' y) (\<lambda>_ _. P)"
  shows "ovalidNF (\<lambda>_. unat x < measure') (fib' measure' x) (\<lambda>_ _. True)"
  apply (subst fib'.simps)
  apply (wp ind)
  apply simp
  apply unat_arith
  done
lemma fib'_term: "no_ofail (\<lambda>_. measure' > unat n) (fib' measure' n)"
  apply (subst no_ofail_is_ovalidNF)
  apply (induct n arbitrary: measure' rule: less_induct)
  apply (subgoal_tac "\<And>y P measure'. ovalidNF (\<lambda>_. y < x \<and> unat y < measure' \<and> P) (fib' measure' y) (\<lambda>_ _. P)")
  apply (rule fib'_term_rec_helper, assumption)
  apply (metis ovalidNF_assume_pre)
  done

(* The overall correctness proof. *)
lemma fib'_spec: "ovalidNF (\<lambda>s. P s \<and> m > unat n) (fib' m n) (\<lambda>r s. P s \<and> r = fibo32 n)"
  apply (rule ovalidNF_combine)
  apply (wp fib'_correct, simp)
  apply (wp fib'_term, simp)
  done

(*
If fib\<^bsub>C\<^esub> was lifted to the state monad...
Once we get hoare triples and wp reasoning on the option monad,
the proof might also look like this.

(* isar proof lets us rewrite the inductive assumption r and pass it to wp *)
lemma fib2'_correct: "\<lbrace> \<lambda>_. P (fibo32 n) \<rbrace> fib2' m n \<lbrace> \<lambda>r s. P r \<rbrace>"
proof (induct n arbitrary: P m rule: less_induct)
   fix x P m
   assume r: "\<And>y P m. y < x \<Longrightarrow> \<lbrace>  \<lambda>_. P (fibo32 y) \<rbrace> fib2' m y \<lbrace> \<lambda>r s. P r \<rbrace>"

   have r': "\<And>y P m. \<lbrace>  \<lambda>_. y < x \<and> P (fibo32 y) \<rbrace> fib2' m y \<lbrace> \<lambda>r s. P r \<rbrace>"
    apply (rule hoare_assume_pre)
    apply clarsimp
    apply (insert r)
    apply (clarsimp simp: valid_def)
    done

   show "\<lbrace> \<lambda>_. P (fibo32 x) \<rbrace> fib2' m x \<lbrace> \<lambda>r s. P r \<rbrace>"
     apply (subst fib2'.simps)
     apply (wp r')
    apply auto
    apply (simp_all add: fibo32.simps)
    apply unat_arith+
    done
qed
*)


(* To be nice to anyone who calls fib', we can define its measure function now. *)
local_defs measure_of'_fib: "measure_of'_fib n \<equiv> \<lambda>_. unat n + 1"

(* Then we can show that anyone who calls fib' will get the correct result. *)
lemma fib'_call:
  "ovalidNF (\<lambda>s. P s \<and> measure' = measure_of'_fib n s)
            (fib' measure' n)
            (\<lambda>r s. P s \<and> r = fibo32 n)"
  unfolding measure_of'_fib
  apply (wp fib'_spec)
  apply simp
  done

lemma "ovalidNF P call_fib' (\<lambda>_. P)"
  apply (unfold call_fib'_def)
  apply wp
   (* FIXME: get wp to do this *)
   apply (rule_tac P = "\<lambda>s. P s \<and> r = measure_of'_fib 0x2A s" in ovalidNF_post_imp_assuming_pre)
    apply simp
   apply simp
   apply (rule fib'_call)
  apply wp
  apply simp
  done

end (* context fib *)

end
