(*
 * Copyright (C) 2013 NICTA
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer,
 *    without modification.
 *
 * 2. Redistributions in binary form must reproduce at minimum a disclaimer
 *    substantially similar to the "NO WARRANTY" disclaimer below
 *    ("Disclaimer") and any redistribution must be conditioned upon
 *    including a substantially similar Disclaimer requirement for further
 *    binary redistribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGES.
 *)

(* interface to Norbert Schirmer's Hoare Package code *)
structure HPInter =
struct

type fninfo =
     {fname : string,
      inparams : (string * typ * int Absyn.ctype option) list,
      outparams : (string * typ * int Absyn.ctype) list,
      body : Absyn.ext_decl HoarePackage.bodykind option,
      spec : (string * Element.context) list,
      mod_annotated_protop : bool,
      recursivep : bool}




(* make function information - perform a preliminary analysis on the
   abstract syntax to extract function names, parameters and specifications *)
fun mk_fninfo thy csenv decllist : fninfo list = let
  open Absyn NameGeneration
  fun parse_spec fname fspec =
      case fspec of
        fnspec spec => let
          val toklist = Outer_Syntax.scan Position.none (node spec)
                        (* it might be nice to remember where the spec came
                           from so we could put in a reasonable pos'n above *)
          val filtered = List.filter Token.is_proper toklist
          val eof = Token.stopper
          val nameproplist = valOf (Scan.read eof
                                              HoarePackage.proc_specs
                                              filtered)
                             handle Option =>
                               raise Fail ("Failed to scan spec for "^fname)
          fun nprop2Assume (name, prop) =
              ((Binding.name name, []), [(prop, [])])
        in
          [("", Element.Assumes (map nprop2Assume nameproplist))]
        end
      | fn_modifies slist => let
          val mgoal = Modifies_Proofs.gen_modify_goalstring csenv fname slist
        in
          [("", Element.Assumes [((Binding.name (fname ^ "_modifies"), []),
                                  [(mgoal, [])])])]
        end
      | didnt_translate => []
      | gcc_attribs _ => []
      | relspec _ => [] (* fixme *)
  fun parse_specs fname fspecs =
      List.concat (map (parse_spec fname) fspecs)
  fun mk_ret rettype =
      if rettype = Void then []
      else [(return_var_name rettype,
             CalculateState.ctype_to_typ (thy, rettype),
             rettype)]
  fun to_param vi = let
    open CalculateState ProgramAnalysis
    val cty = get_vi_type vi
    val typ = ctype_to_typ (thy,cty)
  in
    (get_mname vi, typ, SOME cty)
  end
  fun build_tab (d,tab) =
      case d of
        FnDefn((_, fname), _, _, _) => Symtab.update(node fname,d) tab
      | Decl d0 => let
        in
          case node d0 of
            ExtFnDecl(ed0 as {name = name_w,...}) => let
              val name = node name_w
            in
              case Symtab.lookup tab name of
                NONE => Symtab.update(name, d) tab
              | SOME (FnDefn _) => tab
              | SOME (Decl d1) => let
                in
                  case node d1 of
                    ExtFnDecl {specs,name,rettype,params} => let
                      val newd0 = {rettype=rettype,name=name,params=params,
                                   specs=merge_specs (#specs ed0) specs}
                      val newd = wrap(ExtFnDecl newd0, left d0, right d0)
                    in
                      Symtab.update (node name, Decl newd) tab
                    end
                  | _ => tab
                end
            end
          | _ => tab
        end
  val table = List.foldl build_tab Symtab.empty decllist
  fun foldthis (_, d) acc =
      case d of
        FnDefn((rettype0,fname_w), inparams, prepost, body) => let
          open HoarePackage CalculateState ProgramAnalysis
          val fname = node fname_w
          val rettype = valOf (get_rettype fname csenv)
        in
          {fname = fname,
           inparams = map to_param (valOf (get_params fname csenv))
                      handle Option =>
                        raise Fail ("mk_fninfo: No function information for "^
                                    fname),
           outparams = mk_ret rettype,
           recursivep = is_recursivefn csenv fname,
           body = SOME (BodyTerm d), spec = parse_specs fname prepost,
           mod_annotated_protop = false}
          ::
          acc
        end
      | Decl d0 => let
        in
          case node d0 of
            ExtFnDecl { name, specs, params = pms0, rettype = rettype0 } => let
              open CalculateState ProgramAnalysis
              val fname = node name
              val params = map to_param (valOf (get_params fname csenv))
                  handle Option =>
                         raise Fail ("mk_fninfo: No function information for "^
                                     fname)
              val rettype = valOf (get_rettype fname csenv)
	      val _ = writeln ("Added external decl for " ^ fname ^
			       " with " ^ Int.toString (length params) ^
                               " args.")
              val spec = parse_specs fname specs
              val mod_annotated_protop = isSome (get_modifies csenv fname)
              val body = NONE
            in
              {fname = fname,
               inparams = params, recursivep = false,
               outparams = mk_ret rettype, body = body,
               spec = spec, mod_annotated_protop = mod_annotated_protop}
              ::
              acc
            end
          | _ => acc
        end
      | _ => acc
in
  Symtab.fold foldthis table []
end

fun mkSpecFunction thy cse styargs (f : fninfo) = let
  open TermsTypes
  val {fname,...} = f
  val state_ty = hd styargs
  val mods = fname |> ProgramAnalysis.get_modifies cse
                   |> valOf |> Binaryset.listItems
  val s = Free("s", state_ty)
  val t = Free("t", state_ty)
  val relbody = Modifies_Proofs.gen_modify_body thy state_ty s t mods
  val collect_t = mk_collect_t (mk_prod_ty (state_ty, state_ty))
  val fbody = mk_Spec(styargs, collect_t $ list_mk_pabs([s,t], relbody))
in
  (fname, fbody, fbody)
end


(* compile bodies - turn a list of AST values into a term list *)
fun compile_fninfos cse styargs compfn lthy (fninfo : fninfo list) = let
  open HoarePackage
  fun partition (acc as (asts,specs)) fns =
      case fns of
        [] => (List.rev asts, specs)
      | {body = SOME (BodyTerm x),...} :: t => partition (x::asts,specs) t
      | (i as {mod_annotated_protop = true, ...}) :: t =>
          partition (asts,i::specs) t
      | _ :: t => partition acc t
  val (asts,toSpec) = partition ([],[]) fninfo
in
  compfn asts lthy @
  map (mkSpecFunction (Proof_Context.theory_of lthy) cse styargs) toSpec
end

fun rhs_extras rhs =
    Term.fold_aterms
        (fn v as Free (x, _) => insert (op aconv) v | _ => I)
        rhs
        [];

fun mk_fun_types [] T = T
  | mk_fun_types (Free (X, T')::ts) T = T' --> mk_fun_types ts T

fun extract_fixes elems = let
  open Element
  fun fix2term (n,tyopt,_) =
      case tyopt of
        NONE => NONE
      | SOME ty => SOME (Free(Binding.name_of n,ty))
  fun recurse elems =
      case elems of
        [] => []
      | Fixes fs::rest => List.mapPartial fix2term fs @ recurse rest
      | _ :: rest => recurse rest
in
  recurse elems
end

fun create_bodyloc_elems fninfo globs thy (name, body_t, body_stripped_t) =
let
  val rhs_vars0 = rhs_extras body_t

  (* Library.subtract has arguments in wrong order - sheesh! *)
  val rhs_vars = subtract (op aconv) globs rhs_vars0

  fun mk_rhs(nm, vars, t) = let
    val t0 = TermsTypes.list_mk_abs (vars, t)
  in
    (nm, Term.map_types (Sign.certify_typ thy) t0)
  end

  val eqt_stripped = mk_rhs(name ^ "_body", [], body_stripped_t)
  val eqt = mk_rhs(name ^ (if null rhs_vars then "_body" else "_invs_body"),
                   rhs_vars, body_t)
  fun eqn_to_definfo (nm, t) = let
    val b = Binding.name nm
  in
    ((b, NoSyn), ((Thm.def_binding b, []), t))
  end
  val elems =
      if null rhs_vars then [eqn_to_definfo eqt]
      else [eqn_to_definfo eqt_stripped, eqn_to_definfo eqt]
  type lthy_def_info = (binding * mixfix) * (Attrib.binding * term)
in
  (elems : lthy_def_info list,
   (name, if null rhs_vars then body_t else body_stripped_t))
end

val globalsN = "_global_addresses"
fun add_syntax (name,recp,inpars,outpars) thy = let
  open HoarePackage
  val name = suffix HoarePackage.proc_deco name
  val pars = map (fn par => (In,varname par)) inpars@
             map (fn par => (Out,varname par)) outpars

  val thy_decl =
      thy |> Context.theory_map (add_params Morphism.identity name pars)
          |> Context.theory_map
              (add_state_kind Morphism.identity name HoarePackage.Record)
          |> (if recp then
                Context.theory_map (add_recursive Morphism.identity name)
              else (fn x => x))
in
  thy_decl
end

fun asm_to_string tmw e =
    case e of
      Element.Assumes asms => let
        fun asm_element pfx ((nm, attrs), ttmlist) =
            pfx ^ Binding.name_of nm ^ ": " ^
            commas (map (tmw o #1) ttmlist) ^ "\n"
      in
        "Assumes:" ^ (if length asms = 1 then
                        " " ^ String.concat (map (asm_element "") asms)
                      else
                        "\n" ^ String.concat (map (asm_element "  ") asms))
      end
    | Element.Fixes stysynlist =>
        "Fixes: " ^ commas (map (Binding.name_of o #1) stysynlist) ^ "\n"
    | _ => "??\n"
;

fun extract_element_asms e =
    case e of
      Element.Assumes asms => map (fn (nm, [(t,_)]) => (nm, t)) asms
    | _ => raise Fail "extract_element_term: malformed element"
;

(* The following is modelled on the code in HoarePackage.procedures_definition
*)
fun make_function_definitions localename
                              (cse : ProgramAnalysis.csenv)
                              (styargs : typ list)
                              (nmdefs : thm list)
                              (fninfo : fninfo list)
                              compfn
                              globloc
                              globals_elems
                              thy =
let
  open HoarePackage
  val localename_b = Binding.name localename
  val thy = thy |> Context.theory_map (HoarePackage.set_default_state_kind
                                       HoarePackage.Record)
                |> Theory.checkpoint
                (* dunno why the checkpoint; just copying NS's code *)
  val writeln = fn s => writeln("HPInter.mfd: " ^ s)
  val name_pars =
      map (fn {fname,inparams,outparams,recursivep,...} =>
                (fname, recursivep, map #1 inparams, map #1 outparams))
          fninfo
  val name_specs = map (fn {fname,spec,...} => (fname,spec)) fninfo
  val thy = List.foldr (uncurry add_syntax) thy name_pars
  val lthy = Named_Target.init I globloc thy
  val name_body = compile_fninfos cse styargs compfn lthy fninfo
  val _ = writeln "Translated all functions"
  val termwriter = Syntax.string_of_term lthy
  val globs = extract_fixes globals_elems

  val body_elems = map (create_bodyloc_elems fninfo globs thy) name_body

  fun add_body_defs ((elems, (name,bod)), lthy) = let
    fun foldthis (e, lthy) = let
      val _ = writeln ("Adding body_def for " ^ Binding.name_of (#1 (#1 e)))
    in
      Local_Theory.restore (#2 (Local_Theory.define e lthy))
    end
  in
    List.foldl foldthis lthy elems
  end
  val lthy = List.foldl add_body_defs lthy body_elems
  val lthy = Local_Theory.exit lthy
  val ctxt = lthy

  (* set up _impl by defining \<Gamma> *)
  val (defs, ctxt) =
      if List.exists (isSome o #body) fninfo then let
          val lthy =
              Named_Target.init I globloc (Proof_Context.theory_of ctxt)
          fun expfname fni = let
            val nm = #fname fni
          in
            case (#body fni, #mod_annotated_protop fni) of
              (NONE, false) => NONE
            | _ => let
                val realnm =
                    Consts.intern (Proof_Context.consts_of lthy) (nm ^ "_body")
              in
                SOME (Syntax.check_term lthy (Const(realnm, dummyT)))
              end
          end
          val blist' = map expfname fninfo
          val bnm = ListPair.zip (blist',nmdefs)
          val triples =
              ListPair.zip (bnm,
                            map (suffix HoarePackage.implementationN o #fname)
                                fninfo)
          val (defs, lthy) =
              StaticFun.define_tree_and_thms "\<Gamma>" triples lthy
          val lthy' = Local_Theory.restore lthy
          val morph = Proof_Context.export_morphism lthy lthy'
          val defs' = map (fn (n, t) =>
                              (Morphism.binding morph (Binding.name n),
                               map (Morphism.thm morph) t))
                          defs
        in
          (defs', Local_Theory.exit lthy')
        end
      else ([], ctxt)

  val globloc_expr = ([(globloc, (("", false), Expression.Named []))], [])

  (* three levels of folding - woohoo *)
  fun add_spec_locales ((name,specs), (localemap, ctxt)) = let
    (* add any spec locales *)
    fun foldthis ((_, e), (localemap, ctxt)) = let
      val lnm_strs = extract_element_asms e
      fun innerfold (((speclocalename_b,_), tstr), (localemap, ctxt)) = let
        val speclocalename = Binding.name_of speclocalename_b
        val thy = Proof_Context.theory_of ctxt
        val t = Syntax.read_term ctxt tstr
        val _ = writeln ("Adding spec locale "^speclocalename^
                         (if speclocalename = "" then "" else " ")^
                         "for function "^name ^ " (" ^ tstr ^ ")")
        val e' = Element.Assumes [((speclocalename_b,[]),
                                   [(TermsTypes.mk_prop t, [])])]
        val speclocalename_b = Binding.map_name
                                   (fn s => if s = "" then
                                              suffix HoarePackage.specL name
                                            else s)
                                   speclocalename_b
        val (locname, ctxt') =
            Expression.add_locale I
                speclocalename_b
                speclocalename_b
                globloc_expr
                [e']
                thy
        (* localename is the name we wanted, locname is the name Isabelle
           gives back to us.  This will be properly qualified yadda yadda *)
      in
        (Symtab.update (speclocalename, locname) localemap, ctxt')
      end
    in
      List.foldl innerfold (localemap,ctxt) lnm_strs
    end
    val (speclocnames,ctxt) = List.foldl foldthis (localemap, ctxt) specs
  in
    (speclocnames, ctxt)
  end
  val (specloc_names, ctxt) =
      List.foldl add_spec_locales (Symtab.empty, ctxt) name_specs

  val (globloc, ctxt) =
      Expression.add_locale I localename_b localename_b globloc_expr []
                            (Proof_Context.theory_of ctxt)

in
  (globloc, defs, Proof_Context.theory_of ctxt)
end


end (* struct *)
