(*
 * Copyright (C) 2013, National ICT Australia Limited. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *  * The name of National ICT Australia Limited nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *)

theory WhileLoopRules
imports "wp/NonDetMonadVCG"
begin

section {* Rules about whileLoops. *}

text {*
  The following rules re-interpret "while" loops as a directed
  graph. The graph contains standard nodes and "final" nodes,
  the latter which correspond to the while loop's condition
  becoming false. Final nodes may not have outgoing edges from
  them.
*}

definition "whileLoop_results_step C B x y \<equiv>
  case x of
    None \<Rightarrow> False
  | Some (r, s) \<Rightarrow>
       C r s \<and> (y \<in> Some ` fst (B r s) \<or> (y = None \<and> snd (B r s)))"

lemma whileLoop_results_step_simps [simp]:
  "\<not> whileLoop_results_step C B None z"
  "whileLoop_results_step C B (Some x) (Some y)
       = (C (fst x) (snd x) \<and> y \<in> fst (B (fst x) (snd x)))"
  "whileLoop_results_step C B (Some x) None
       = (C (fst x) (snd x) \<and> snd (B (fst x) (snd x)))"
  by (auto simp: whileLoop_results_step_def)

definition "whileLoop_results_final_node C B x \<equiv>
   case x of
     None \<Rightarrow> True
   | Some (r, s) \<Rightarrow> \<not> C r s"

lemma whileLoop_results_final_node_simps [simp]:
  "whileLoop_results_final_node C B None"
  "whileLoop_results_final_node C B (Some x) = (\<not> C (fst x) (snd x))"
  by (auto simp: whileLoop_results_final_node_def split_def)

lemma whileLoop_no_step_from_final:
  "\<lbrakk> whileLoop_results_step C B x y \<rbrakk> \<Longrightarrow> \<not> whileLoop_results_final_node C B x"
  by (clarsimp simp: whileLoop_results_step_def whileLoop_results_final_node_def split: option.splits)

definition "whileLoop_results_graph C B \<equiv> {(x, y). whileLoop_results_step C B x y }"

lemma while_to_rtrancl':
  shows "((x, y) \<in> whileLoop_results C B) =
                ((x, y) \<in> rtrancl (whileLoop_results_graph C B)
                         \<and> whileLoop_results_final_node C B y)"
     (is "?LHS x y = ?RHS x y")
proof (rule iffI)
  assume "?RHS x y"
  thus "?LHS x y"
    apply clarsimp
    apply (induct rule: converse_rtrancl_induct)
     apply (case_tac y, auto)[1]
    apply clarsimp
    apply (subst (asm) whileLoop_results_graph_def)
    apply (clarsimp simp: whileLoop_results_step_def image_def split: option.splits)
    apply (auto intro: whileLoop_results.intros)
    done
next
  assume "?LHS x y"
  thus "?RHS x y"
    apply (induct rule: whileLoop_results.induct)
       apply simp
      apply clarsimp
     apply clarsimp
     apply (rule r_into_rtrancl)
     apply (clarsimp simp: whileLoop_results_graph_def)
    apply clarsimp
    apply (erule converse_rtrancl_into_rtrancl [rotated])
    apply (clarsimp simp: whileLoop_results_graph_def)
    done
qed

lemma while_to_rtrancl:
  shows "(whileLoop_results C B) =
                {(x, y). ((x, y) \<in> (rtrancl (whileLoop_results_graph C B))
                         \<and> (whileLoop_results_final_node C B y))}"
  using while_to_rtrancl'
  by blast

subsection "Simple base rules"

lemma not_snd_loop_terminates:
  "\<not> snd (whileLoop C B r s) \<Longrightarrow> whileLoop_terminates C B r s"
  apply (clarsimp simp: whileLoop_def)
  done

lemma not_snd_loop_terminatesE:
  "\<not> snd (whileLoopE C B r s) \<Longrightarrow> whileLoop_terminatesE C B r s"
  apply (clarsimp simp: whileLoopE_def whileLoop_terminatesE_def)
  apply (rule not_snd_loop_terminates)
  apply simp
  done

lemma whileLoop_results_cong_lemma:
  "\<lbrakk> (x, y) \<in> whileLoop_results C B; C = C'; \<And>r s. C' r s \<Longrightarrow> B r s = B' r s \<rbrakk>
      \<Longrightarrow> (x, y) \<in> whileLoop_results C' B'"
  apply (induct rule: whileLoop_results.induct)
     apply clarsimp
    apply clarsimp
   apply (rule whileLoop_results.intros, auto)[1]
  apply clarsimp
  apply (rule whileLoop_results.intros, auto)[1]
  done

lemma whileLoop_results_cong [cong]:
  "\<lbrakk> \<And>r s. C r s = C' r s; \<And>r s. C' r s \<Longrightarrow> B r s = B' r s \<rbrakk>
       \<Longrightarrow> whileLoop_results C B = whileLoop_results C' B'"
  apply (rule set_eqI, rule iffI)
   apply (case_tac x, simp)
   apply (erule whileLoop_results_cong_lemma)
    apply (rule ext, rule ext, simp)
   apply clarsimp
  apply (case_tac x, simp)
  apply (erule whileLoop_results_cong_lemma)
   apply (rule ext, rule ext, simp)
  apply simp
  done

subsection "Basic induction helper lemmas"

lemma whileLoop_rule_lemma1:
    "\<lbrakk> (a, b) \<in> whileLoop_results C B; b = Some (x, y) \<rbrakk> \<Longrightarrow> \<not> C x y"
    apply (induct rule: whileLoop_results.induct, auto)
    done

lemma whileLoop_rule_lemma1':
    "\<lbrakk> (a, b) \<in> whileLoop_results C B; a \<noteq> b \<rbrakk> \<Longrightarrow> \<exists>x. a = Some x \<and> C (fst x) (snd x)"
    apply (induct rule: whileLoop_results.induct, auto)
    done

lemma whileLoop_rule_lemma2 [consumes 1]:
  "\<lbrakk> (a, b) \<in> whileLoop_results C B;
     a = Some x; b = Some y;
     P x; \<And>s t. \<lbrakk> P s; t \<in> fst (B (fst s) (snd s)); C (fst s) (snd s) \<rbrakk> \<Longrightarrow> P t \<rbrakk> \<Longrightarrow> P y"
  apply (induct arbitrary: x y rule: whileLoop_results.induct)
     apply simp
    apply simp
   apply simp
  apply atomize
  apply fastforce
  done

lemma whileLoop_rule_lemma3 [consumes 1]:
  assumes result: "(Some (r, s), Some (r', s')) \<in> whileLoop_results C B"
  and inv_start: "I r s"
  and inv_step: "\<And>r s r' s'. \<lbrakk> I r s; C r s; (r', s') \<in> fst (B r s) \<rbrakk> \<Longrightarrow> I r' s'"
  shows "I r' s'"
  apply (rule whileLoop_rule_lemma2 [where P="split I" and y="(r', s')" and x="(r, s)",
      simplified split_def, simplified])
      apply (rule result)
     apply simp
    apply simp
   apply fact
  apply (erule (1) inv_step)
  apply clarsimp
  done

lemma whileLoop_results_induct_lemma [consumes 1]:
  "\<lbrakk>(x1, x2) \<in> whileLoop_results C B; x1 = Some (r, s); x2 = Some (r', s');
    \<And>r s. \<not> C r s \<Longrightarrow> P r s r s;
    \<And>r s r' s' r'' s''. \<lbrakk>C r s; (r', s') \<in> fst (B r s);
             (Some (r', s'), Some (r'', s'')) \<in> whileLoop_results C B; P r' s' r'' s'' \<rbrakk> \<Longrightarrow> P r s r'' s''
   \<rbrakk> \<Longrightarrow> P r s r' s'"
  apply (induct arbitrary: r s r' s' rule: whileLoop_results.induct)
     apply simp
    apply clarsimp
   apply clarsimp
  apply clarsimp
  done

lemma whileLoop_results_induct [consumes 1]:
  "\<lbrakk>(Some (r, s), Some (r', s')) \<in> whileLoop_results C B;
    \<And>r s. \<not> C r s \<Longrightarrow> P r s r s;
    \<And>r s r' s' r'' s''. \<lbrakk>C r s; (r', s') \<in> fst (B r s);
             (Some (r', s'), Some (r'', s'')) \<in> whileLoop_results C B; P r' s' r'' s'' \<rbrakk> \<Longrightarrow> P r s r'' s''
   \<rbrakk> \<Longrightarrow> P r s r' s'"
  by (erule whileLoop_results_induct_lemma, auto)

lemma whileLoop_results_induct_fail [consumes 1]:
  assumes failed_result: "(Some (r, s), None) \<in> whileLoop_results C B"
  and finish: "\<And>r s. \<lbrakk> C r s; snd (B r s) \<rbrakk> \<Longrightarrow> P r s"
  and step: "\<And>r s r' s'. \<lbrakk>C r s; (r', s') \<in> fst (B r s);
             (Some (r', s'), None) \<in> whileLoop_results C B; P r' s' \<rbrakk> \<Longrightarrow> P r s"
  shows "P r s"
proof -
  {
    fix a b
    assume x: "(a, b) \<in> whileLoop_results C B"
    and y: "a = Some (r, s)"
    and z: "b = None"
    have "(\<lambda>a b. P (fst (the a)) (snd (the a))) a b"
      using x y z
      apply (induct arbitrary: r s rule: whileLoop_results.induct)
         apply simp
        apply clarsimp
       apply clarsimp
       apply (erule (1) finish)
      apply clarsimp
      apply (erule step, auto)[1]
      done
  }

  hence "P (fst (the (Some (r, s)))) (snd (the (Some (r, s))))"
    apply this
    apply (rule failed_result, auto)
    done

  thus ?thesis
    by simp
qed

lemma whileLoop_induct [consumes 1]:
  "\<lbrakk> (r', s') \<in> fst (whileLoop C B r s);
     \<And> r s. \<not> C r s \<Longrightarrow> I r s r s;
     \<And>r s r' s' r'' s''.
        \<lbrakk> C r s; (r', s') \<in> fst (B r s);
          (r'', s'') \<in> fst (whileLoop C B r' s');
           I r' s' r'' s'' \<rbrakk> \<Longrightarrow> I r s r'' s''
   \<rbrakk> \<Longrightarrow> I r s r' s'"
  apply atomize
  apply (clarsimp simp: whileLoop_def)
  apply (induct rule: whileLoop_results_induct_lemma, simp_all)
  done

(* Rule for reasoning about 'whileLoop_terminates', and a proof
 * that the rule is complete. *)

lemma not_whileLoop_terminates:
  assumes init_I: "I r s"
      and cond_I: "C r s"
      and non_term:  "\<And>r s r' s'. \<lbrakk> I r s; C r s \<rbrakk> \<Longrightarrow> \<exists>(r', s') \<in> fst (B r s). C r' s' \<and> I r' s'"
      shows "\<not> whileLoop_terminates C B r s"
  apply (rule notI)
  apply (insert init_I cond_I)
  apply (induct rule: whileLoop_terminates.induct)
   apply clarsimp
  apply clarsimp
  apply (frule (1) non_term)
  apply (clarsimp simp: Bex_def)
  apply force
  done

lemma not_whileLoop_terminates_complete:
  shows "(\<not> whileLoop_terminates C B r s) =
              (\<exists>I. I r s \<and> C r s \<and>
                   (\<forall>r s. (I r s \<and> C r s) \<longrightarrow> (\<exists>(r', s') \<in> fst (B r s). C r' s' \<and> I r' s')))"
    (is "?LHS = ?RHS")
proof (rule iffI)
  assume "?LHS"
  thus "?RHS"
    apply -
    apply (rule exI [where x="\<lambda>r s. \<not> whileLoop_terminates C B r s"])
    apply clarsimp
    apply (rule conjI)
     apply (subst (asm) whileLoop_terminates.simps)
     apply clarsimp
    apply clarsimp
    apply (subst (asm) (2) whileLoop_terminates.simps)
    apply clarsimp
    apply (erule bexI [rotated])
    apply clarsimp
    apply (subst (asm) (2) whileLoop_terminates.simps)
    apply simp
    done
next
  assume "?RHS"
  thus "?LHS"
    apply clarsimp
    apply (cut_tac C=C and B=B and I=I and r=r and s=s in not_whileLoop_terminates)
    apply auto
    done
qed

(* A rule for reasoning about failure and non-failure, and also
 * proofs of their completeness. *)

lemma snd_whileLoop:
  assumes init_I: "I r s"
      and cond_I: "C r s"
      and non_term:  "\<And>r s. \<lbrakk> I r s; C r s; \<not> snd (B r s) \<rbrakk>
                            \<Longrightarrow> \<exists>(r', s') \<in> fst (B r s). C r' s' \<and> I r' s'"
      shows "snd (whileLoop C B r s)"
  apply (clarsimp simp: whileLoop_def)
  apply (rotate_tac)
  apply (insert init_I cond_I)
  apply (induct rule: whileLoop_terminates.induct)
  apply clarsimp
  apply (frule (1) non_term)
   apply (subst (asm) (2) whileLoop_results.simps)
   apply clarsimp
  apply clarsimp
  apply (insert whileLoop_results.simps)
  apply fast
  done

lemma snd_whileLoop_complete:
  shows "snd (whileLoop C B r s) =
              (\<exists>I. I r s \<and> C r s \<and>
                   (\<forall>r s. (I r s \<and> C r s \<and> \<not> snd (B r s)) \<longrightarrow>
                            (\<exists>(r', s') \<in> fst (B r s). C r' s' \<and> I r' s')))"
  (is "?LHS = ?RHS")
proof (rule iffI)
  assume "?LHS"
  thus "?RHS"
    apply (clarsimp simp: whileLoop_def)
    apply (erule disjE)
     apply (rule exI [where x="\<lambda>r s. (Some (r, s), None) \<in> whileLoop_results C B"])
     apply clarsimp
     apply (drule whileLoop_rule_lemma1')
      apply simp
     apply clarsimp
     apply (erule whileLoop_results_cases_fail, simp)
     apply clarsimp
     apply (erule bexI [rotated])
     apply clarsimp
     apply (drule whileLoop_rule_lemma1')
      apply simp
     apply clarsimp
    apply (rule exI [where x="\<lambda>r s. \<not> whileLoop_terminates C B r s"])
    apply clarsimp
    apply (rule conjI)
     apply (subst (asm)  whileLoop_terminates_simps)
     apply clarsimp
    apply clarsimp
    apply (subst (asm) (2)  whileLoop_terminates_simps)
    apply clarsimp
    apply (erule bexI [rotated])
    apply clarsimp
    apply (subst (asm) (2)  whileLoop_terminates_simps)
    apply clarsimp
    done
next
  assume "?RHS"
  thus "?LHS"
    apply clarsimp
    apply (erule (1) snd_whileLoop)
    apply force
    done
qed

lemma not_snd_whileLoop:
  assumes init_I: "I r s"
      and wf_R: "wf R"
      and inv_holds:  "\<And>r s r' s'. \<lbrakk> I r s; C r s; (r', s') \<in> fst (B r s) \<rbrakk> \<Longrightarrow> I r' s'"
      and no_fail:  "\<And>r s. \<lbrakk> I r s; C r s \<rbrakk> \<Longrightarrow> \<not> snd (B r s)"
      and R_holds: "\<And>r s r' s'. \<lbrakk> I r s; C r s; (r', s') \<in> fst (B r s) \<rbrakk> \<Longrightarrow> ((r', s'), (r, s)) \<in> R"
      shows "\<not> snd (whileLoop C B r s)"
proof -
  have "\<not> (Some (r, s), None) \<in> whileLoop_results C B"
    apply (rule notI)
    apply (insert init_I)
    apply (induct rule: whileLoop_results_induct_fail)
     apply (metis no_fail)
    apply (metis inv_holds)
    done

  also have "whileLoop_terminates C B r s"
     apply (insert init_I)
     using wf_R
     apply (induction "(r, s)" arbitrary: r s)
     apply atomize
     apply (subst whileLoop_terminates_simps)
     apply clarsimp
     apply (frule (2) R_holds)
     apply (metis inv_holds)
     done

  ultimately show ?thesis
     by (clarsimp simp: whileLoop_def)
qed

lemma whileLoop_terminates_wf:
    "wf {(x, y). C (fst y) (snd y) \<and> x \<in> fst (B (fst y) (snd y)) \<and> whileLoop_terminates C B (fst y) (snd y)}"
  apply (rule wfI [where A="UNIV" and B="{(r, s). whileLoop_terminates C B r s}"])
   apply clarsimp
  apply clarsimp
  apply (erule whileLoop_terminates.induct)
   apply blast
  apply blast
  done

lemma not_snd_whileLoop_complete:
  shows "(\<not> snd (whileLoop C B r s)) = (\<exists>I R . I r s \<and> wf R \<and>
           (\<forall>r s r' s'. (I r s \<and> C r s \<and> (r', s') \<in> fst (B r s)) \<longrightarrow> I r' s')
           \<and> (\<forall>r s. I r s \<and> C r s \<longrightarrow> \<not> snd (B r s))
           \<and> (\<forall>r s r' s'. I r s \<and> C r s \<and> (r', s') \<in> fst (B r s) \<longrightarrow> ((r', s'), (r, s)) \<in> R))"
   (is "?LHS = ?RHS")
proof (rule iffI)
  assume "?RHS"
  thus "?LHS"
    apply (clarsimp)
    apply (erule make_pos_goal, rule not_snd_whileLoop)
    apply (assumption | blast)+
    done
next
  assume no_fail: "?LHS"
  def I \<equiv> "\<lambda>r s. \<not> snd (whileLoop C B r s)"
  def R \<equiv> "{((r', s'), (r, s)). C r s \<and> (r', s') \<in> fst (B r s) \<and> whileLoop_terminates C B r s}"

  have "I r s"
    by (clarsimp simp: I_def no_fail)

  moreover have "wf R"
    apply (rule wf_subset [OF whileLoop_terminates_wf [where C=C and B=B]])
    apply (clarsimp simp: R_def)
    done

  moreover have "\<And>r s r' s'. \<lbrakk> I r s; C r s; (r', s') \<in> fst (B r s) \<rbrakk> \<Longrightarrow> I r' s'"
    apply (clarsimp simp: I_def whileLoop_def)
    apply (rule conjI)
     apply (metis whileLoop_results_simps)
    apply (erule whileLoop_terminates_cases)
     apply clarsimp
    apply fastforce
    done

  moreover have "\<And>r s. \<lbrakk> I r s; C r s \<rbrakk> \<Longrightarrow> \<not> snd (B r s)"
    apply (clarsimp simp: I_def)
    apply (subst (asm) whileLoop_unroll)
    apply (clarsimp simp: condition_true bind_def)
    done

  moreover have "\<And>r s r' s'. \<lbrakk> I r s; C r s; (r', s') \<in> fst (B r s) \<rbrakk> \<Longrightarrow> ((r', s'), (r, s)) \<in> R"
    apply (clarsimp simp: R_def)
    apply (metis I_def not_snd_loop_terminates)
    done

  ultimately show ?RHS
    by metis
qed

(* Weak rule for reasoning about the results of a whileLoop. In particular,
 * it shows the upper bound of the results. *)

lemma fst_whileLoop:
  assumes init_I: "I r s"
    and iter_I:  "\<And>r s r' s'. \<lbrakk> I r s; C r s; (r', s') \<in> fst (B r s) \<rbrakk>
                        \<Longrightarrow> I r' s'"
    and end_Q: "\<And>r s t R. \<lbrakk> \<forall>(r, s) \<in> R. I r s \<and> \<not> C r s \<rbrakk> \<Longrightarrow> Q R"
  shows "Q (fst (whileLoop C B r s))"
  apply (clarsimp simp: whileLoop_def)
  apply (rule end_Q)
  apply clarsimp
  apply (rule conjI)
   apply (erule whileLoop_rule_lemma3, rule init_I, metis iter_I)
  apply (erule whileLoop_rule_lemma1, simp)
  done

lemma whileLoop_rule':
  assumes init_I: "I r s"
      and iter_I:  "\<And>r s r' s'. \<lbrakk> I r s; C r s; (r', s') \<in> fst (B r s) \<rbrakk> \<Longrightarrow> I r' s'"
      and end_Q: "\<And>r s t V f. \<lbrakk> \<forall>(r, s) \<in> fst V. I r s \<and> \<not> C r s \<rbrakk> \<Longrightarrow> Q V"
      shows "Q (whileLoop C B r s)"
  apply (cut_tac fst_whileLoop [where Q="\<lambda>x. Q (x, snd (whileLoop C B r s))", where I=I and r=r and s=s and C=C and B=B])
  apply (auto intro: init_I iter_I intro!: end_Q)
  done

lemma whileLoop_rule_weak:
  assumes init_I: "I r s"
      and iter_I:  "\<And>r. \<lbrace> \<lambda>s. I r s \<and> C r s \<rbrace> B r \<lbrace> \<lambda>r s. I r s \<rbrace>"
      and end_Q: "\<And>r s t V f. \<lbrakk> \<forall>(r, s) \<in> fst V. I r s \<and> \<not> C r s \<rbrakk> \<Longrightarrow> Q V"
      shows "Q (whileLoop C B r s)"
  apply (rule whileLoop_rule' [where I=I and B=B and C=C, OF init_I _ end_Q])
   apply (insert iter_I [unfolded valid_def])[1]
   apply fastforce
  apply simp
  done

lemma whileLoopE_rule_weak:
  assumes init_I: "I r s"
    and iter_I:  "\<And>r. \<lbrace> \<lambda>s. (I r s) \<and> (C r s) \<rbrace> B r \<lbrace> \<lambda>r s. I r s \<rbrace>, \<lbrace> \<lambda>r s. E r s \<rbrace>"
    and end_Q: "\<And>r s t V. \<lbrakk> \<forall>(r', s) \<in> fst V. (\<exists>r. r' = Inr r \<longrightarrow> I r s \<and> \<not> C r s) \<rbrakk> \<Longrightarrow> Q V"
    and except_Q: "\<And>r s t V. \<lbrakk> \<forall>(r', s) \<in> fst V. (\<exists>r. r' = Inl r \<longrightarrow> E r s) \<rbrakk> \<Longrightarrow> Q V"
  shows "Q (whileLoopE C B r s)"
  apply (unfold whileLoopE_def)
  apply (rule whileLoop_rule_weak [
    where Q=Q and I="\<lambda>r s. (case r of Inl x \<Rightarrow> E x s | Inr x \<Rightarrow> I x s)"])
    apply clarsimp
    apply (rule init_I)
   apply (rule hoare_assume_pre)
   apply (clarsimp simp: lift_def split: sum.splits)
   apply (rule hoare_chain [OF iter_I [unfolded validE_def]])
    apply simp
   apply force
  apply (clarsimp split: sum.splits)
  apply (insert end_Q except_Q)
  apply fastforce
  done

lemma fst_whileLoop_cond_false:
  assumes loop_result: "(r', s') \<in> fst (whileLoop C B r s)"
  shows "\<not> C r' s'"
  apply (insert loop_result)
  apply (clarsimp simp: whileLoop_def)
  apply (auto dest!: whileLoop_rule_lemma1)
  done

lemma fst_whileLoop_lower_bound_lemma:
     assumes init_I: "I x"
         and wf_R: "wf R"
        and iter_I: "\<And>x. \<lbrakk> I x; C (fst x) (snd x) \<rbrakk> \<Longrightarrow> \<exists>y. y \<in> fst (B (fst x) (snd x)) \<and> (I y \<and> (y, x) \<in> R)"
        and final_I: "\<And>x. \<lbrakk> I x; \<not> C (fst x) (snd x) \<rbrakk> \<Longrightarrow> x = f"
  shows "f \<in> (fst (whileLoop C B (fst x) (snd x)))"
  apply (insert init_I)
  using wf_R
  apply induction
  apply atomize
  apply (case_tac "C (fst x) (snd x)")
   apply (subst whileLoop_unroll)
   apply (frule (1) iter_I)
   apply (case_tac f)
   apply (clarsimp simp: condition_true in_bind)
   apply force
  apply (frule (1) final_I)
  apply (case_tac f)
  apply (subst whileLoop_unroll)
  apply (clarsimp simp: condition_false in_bind in_return)
  done

lemma fst_whileLoop_lower_bound:
     assumes S_cond: "\<And>r'' s''. \<lbrakk> (r'', s'') \<in> S \<rbrakk> \<Longrightarrow> \<not> C r'' s''"
        and wf_R: "\<And>r'' s''. \<lbrakk> (r'', s'') \<in> S \<rbrakk> \<Longrightarrow> wf (R r'' s'')"
        and init_I: "\<And>r'' s''. \<lbrakk> (r'', s'') \<in> S \<rbrakk> \<Longrightarrow> I r'' s'' r s"
        and iter_I: "\<And>r s r'' s''. \<lbrakk> (r'', s'') \<in> S; I r'' s'' r s; C r s \<rbrakk> \<Longrightarrow>
                      \<exists>(r', s') \<in> fst (B r s). I r'' s'' r' s' \<and> ((r', s'),(r, s)) \<in> (R r'' s'')"
        and final_I: "\<And>r'' s'' r s. \<lbrakk> I r'' s'' r s; \<not> C r'' s''; \<not> C r s \<rbrakk> \<Longrightarrow> (r'' = r \<and> s'' = s)"
  shows "S \<subseteq> (fst (whileLoop C B r s))"
proof (rule subrelI)
  fix f_r f_s
  assume in_set: "(f_r, f_s) \<in> S"
  show "(f_r, f_s) \<in> fst (whileLoop C B r s)"
    apply (rule fst_whileLoop_lower_bound_lemma [
       where x="(r, s)" and I="\<lambda>(r, s). I f_r f_s r s"
       and R="R f_r f_s", simplified])
       apply (rule init_I [OF in_set])
      apply (rule wf_R [OF in_set])
     apply clarsimp
     apply (frule (1) iter_I [OF in_set])
     apply force
    apply clarsimp
    apply (metis S_cond final_I in_set)
    done
qed

lemma fst_whileLoop_lower_bound_terminating:
     assumes S_cond: "\<And>r'' s''. \<lbrakk> (r'', s'') \<in> S \<rbrakk> \<Longrightarrow> \<not> C r'' s''"
        and init_I: "\<And>r'' s''. \<lbrakk> (r'', s'') \<in> S \<rbrakk> \<Longrightarrow> I r'' s'' r s"
        and iter_I: "\<And>r s r'' s''. \<lbrakk> (r'', s'') \<in> S; I r'' s'' r s; C r s \<rbrakk> \<Longrightarrow>
                                 \<exists>(r', s') \<in> fst (B r s). I r'' s'' r' s'"
        and final_I: "\<And>r'' s'' r s. \<lbrakk> I r'' s'' r s; \<not> C r'' s''; \<not> C r s \<rbrakk> \<Longrightarrow> (r'' = r \<and> s'' = s)"
      and terminates: "whileLoop_terminates C B r s"
    shows "S \<subseteq> (fst (whileLoop C B r s))"
  apply (rule fst_whileLoop_lower_bound [
       where I="\<lambda>r' s' r s. I r' s' r s
             \<and> whileLoop_terminates C B r s" and B=B and C=C])
      apply fact
     apply (rule whileLoop_terminates_wf [where B=B and C=C])
    apply (clarsimp simp: terminates init_I)
   apply clarsimp
   apply (frule (2) iter_I)
   apply clarsimp
   apply (rule bexI [rotated], assumption)
   apply clarsimp
   apply (erule whileLoop_terminates.cases, fastforce, fastforce)
  apply (metis final_I)
  done

lemma "fst (whileLoop (\<lambda>(c,r :: nat) _. c)
    (\<lambda>(c, r). do r \<leftarrow> return (r + 1); c \<leftarrow> select UNIV; return (c, r) od)
        (True, 0) s) = {((c', r'), s'). \<not> c' \<and> 0 < r' \<and> s' = s}"
  apply (rule equalityI [rotated])
   apply (rule fst_whileLoop_lower_bound [where
       I="\<lambda>(c', r') t' (c, r) t. \<not> c' \<and> t = t' \<and> (if \<not> c then r = r' else r < r')" and
       R="\<lambda>(_, fr) _. {(((_, r'), _), ((_, r), _)). fr - r' < fr - r}"])
       apply clarsimp
      apply (clarsimp simp: split_def)
      apply (subst in_measure [symmetric], fastforce)
     apply clarsimp
    apply (clarsimp simp: select_def bind_def return_def)
    apply (metis Suc_diff_Suc Suc_lessI lessI)
   apply clarsimp
  apply (rule fst_whileLoop [where I="\<lambda>(c', r') s'. (\<not> c' \<longrightarrow> 0 < r') \<and> s' = s"])
    apply simp
   apply (clarsimp simp: select_def return_def bind_def)
  apply (clarsimp )
  apply force
  done

fun valid_path
where
    "valid_path C B [] = False"
  | "valid_path C B [x] = whileLoop_results_final_node C B x"
  | "valid_path C B (x#y#xs) = (whileLoop_results_step C B x y \<and> valid_path C B (y#xs))"

definition "shortest_path_length C B x y \<equiv>
        (LEAST n. \<exists>l. valid_path C B l \<and> hd l = x \<and> last l = y \<and> length l = n)"

lemma valid_path_simp:
  "valid_path C B l =
        ((\<exists>a. l = [a] \<and> whileLoop_results_final_node C B a \<or>
              (\<exists>x y xs. l = x#y#xs \<and> whileLoop_results_step C B x y \<and> valid_path C B (y#xs))))"
  apply clarsimp
  apply (case_tac l)
   apply clarsimp
  apply (case_tac list)
   apply clarsimp
  apply clarsimp
  done

lemma path_exists:
  assumes path: "(x, y) \<in> whileLoop_results C B"
  shows "\<exists>l. valid_path C B l \<and> hd l = x \<and> last l = y"
  using path
  apply -
  apply (subst (asm) while_to_rtrancl)
  apply clarsimp
  apply (induct rule: converse_rtrancl_induct)
   apply (rule_tac x="[y]" in exI)
   apply clarsimp
  apply clarsimp
  apply (rule_tac x="y#l" in exI)
  apply (case_tac l)
   apply clarsimp
  apply (clarsimp simp: whileLoop_results_graph_def)
  done

lemma shortest_path_exists:
  "(x, y) \<in> whileLoop_results C B \<Longrightarrow>
       \<exists>l. valid_path C B l \<and> shortest_path_length C B x y = length l \<and> hd l = x \<and> last l = y"
  apply (frule path_exists)
  apply (clarsimp simp: simp: shortest_path_length_def)
  apply (rule LeastI2_ex)
   apply force
  apply force
  done

lemma valid_path_length_shorten_path:
  "valid_path C B l \<Longrightarrow> shortest_path_length C B (hd l) (last l) \<le> length l"
  apply (clarsimp simp: simp: shortest_path_length_def)
  apply (rule Least_le)
  apply force
  done

lemma valid_path_None [simp]:
  "valid_path C B (None # xs) = (xs = [])"
  apply (subst valid_path_simp)
  apply clarsimp
  done

lemma valid_path_in_results:
  "valid_path C B l \<Longrightarrow> (hd l, last l) \<in> whileLoop_results C B"
  apply (induct l)
   apply clarsimp
  apply (subst (asm) valid_path_simp)
  apply (clarsimp simp: whileLoop_results_final_node_def
                   whileLoop_results_step_def image_def split: option.splits)
  apply (auto intro: whileLoop_results.intros)
  done

lemma fst_whileLoop_lower_bound_complete:
  shows "(S \<subseteq> (fst (whileLoop C B r s)))
                 = (\<exists>I R. (\<forall>(r'', s'') \<in> S.
                          \<not> C r'' s'' \<and>
                           wf (R r'' s'')
                           \<and> I r'' s'' r s
                           \<and> (\<forall>r s. I r'' s'' r s \<and> C r s \<longrightarrow> (
                                \<exists>(r', s') \<in> fst (B r s). I r'' s'' r' s' \<and> ((r', s'),(r, s)) \<in> R r'' s'')))
                           \<and> (\<forall>r'' s'' r s. I r'' s'' r s \<and> \<not> C r s \<longrightarrow> (r'' = r \<and> s'' = s)))"
        (is "?LHS = (\<exists>I R. (\<forall>(r'', s'')\<in>S. ?RHS r'' s'' I R) \<and> ?final I R)")
proof (rule iffI)
  assume subset: ?LHS
  note subset_results = subset [unfolded whileLoop_def, simplified]

  def I \<equiv> "\<lambda>r'' s'' r s. (Some (r, s), Some (r'', s'')) \<in> whileLoop_results C B"

  def R \<equiv> "\<lambda>r'' s''. measure (\<lambda>x. shortest_path_length C B (Some x) (Some (r'', s'')))"

  show "\<exists>I R. (\<forall>(r'', s'')\<in>S. ?RHS r'' s'' I R) \<and> ?final I R"
  proof (rule exI, rule exI, rule conjI, clarsimp)
    fix r'' s''
    assume r_s_in_set: "(r'', s'') \<in> S"

    have r_s_final: "(Some (r, s), Some (r'', s'')) \<in> whileLoop_results C B"
      using subset_results r_s_in_set
      by blast

    have final_cond_false: "\<not> C r'' s''"
      using fst_whileLoop_cond_false [unfolded whileLoop_def, simplified]
      by (metis r_s_final)

    have init_I: "I r'' s'' r s"
      by (clarsimp simp: I_def r_s_final)

    have inv_holds: "\<And>r s. \<lbrakk> I r'' s'' r s; C r s \<rbrakk> \<Longrightarrow>
                \<exists>(r', s') \<in> fst (B r s). I r'' s'' r' s' \<and> ((r', s'), r, s) \<in> R r'' s'' "
      apply (clarsimp simp: I_def R_def)
      apply (frule shortest_path_exists)
      apply clarsimp
      apply (subst (asm) valid_path_simp)
      apply clarsimp
      apply (erule disjE)
       apply clarsimp
      apply (clarsimp simp: whileLoop_results_step_def image_def)
      apply (case_tac y)
       apply clarsimp
      apply clarsimp
      apply (erule bexI [rotated])
      apply (rule conjI)
       apply (drule valid_path_in_results)
       apply clarsimp
      apply (drule valid_path_length_shorten_path)
      apply clarsimp
      done

    have "wf (R r'' s'')"
       by (clarsimp simp: R_def)

    show "?RHS r'' s'' I R"
       by (clarsimp simp: init_I inv_holds final_cond_false `wf (R r'' s'')`)
  next
    show "\<forall>r'' s'' r s. I r'' s'' r s \<and> \<not> C r s \<longrightarrow> r'' = r \<and> s'' = s"
      apply (clarsimp simp: I_def)
      apply (erule whileLoop_results_cases_valid, auto)
      done
  qed
next
  assume rhs: "\<exists>I R. (\<forall>(r'', s'') \<in> S. ?RHS r'' s'' I R) \<and> ?final I R"
  show ?LHS
    apply (insert rhs)
    apply (erule exE)+
    apply (rule_tac R=R and I=I in fst_whileLoop_lower_bound, fastforce+)
    done
qed

lemma predicate_to_set:
  "\<exists>(R::'a set). (Q W = (W = R))"
  apply (rule exI [where x="(if Q W then W else (- W))"])
  apply clarsimp
  apply (metis Compl_iff)
  done

lemma whileLoop_rule_nofail:
  assumes init_I: "I r s"
    and wf_R: "wf R"
    and iter_I: "\<And>r. \<lbrace> \<lambda>s. I r s \<and> C r s \<rbrace> B r \<lbrace> \<lambda>r s. I r s \<rbrace>!"
    and iter_R: "\<And>r m. \<lbrace>\<lambda>s. I r s \<and> C r s \<and> m = (r, s) \<rbrace> B r \<lbrace> \<lambda>r s. ((r, s), m) \<in> R \<rbrace>"
    and no_fail: "\<And>r s. \<lbrakk> I r s; C r s \<rbrakk> \<Longrightarrow> \<not> snd (B r s)"
    and end_Q: "\<And>r s t V. \<lbrakk> \<forall>(r, s) \<in> fst V. I r s \<and> \<not> C r s; \<not> snd V \<rbrakk> \<Longrightarrow> Q V"
  shows "Q (whileLoop C B r s)"
proof -
  have iter_I':
     "\<And>r s r' s'. \<lbrakk> I r s; C r s; (r', s') \<in> fst (B r s); \<not> snd (B r s) \<rbrakk>
                  \<Longrightarrow> I r' s'"
     by (metis (lifting) use_valid validNF_def iter_I)

  have iter_R':
     "\<And>r s r' s'. \<lbrakk> I r s; C r s; (r', s') \<in> fst (B r s); \<not> snd (B r s) \<rbrakk>
                        \<Longrightarrow> ((r', s'), (r, s)) \<in> R"
    apply (rule use_valid [OF _ iter_R], assumption, clarsimp)
    done


  have "\<not> snd (whileLoop C B r s)"
    apply (rule not_snd_whileLoop [where I=I and R=R and B=B and C=C])
        apply fact+
      apply (rule iter_I', auto simp: no_fail)[1]
     apply fact
    apply (rule iter_R', auto simp: no_fail)[1]
    done

  thus ?thesis
    apply -
    apply (erule rev_mp)
    apply (rule whileLoop_rule' [where I=I, OF init_I])
      apply (rule iter_I', auto simp: no_fail)[1]
    apply (clarsimp simp: Ball_def)
    apply (rule end_Q, auto)[1]
    done
qed

lemma whileLoop_wp:
  "\<lbrakk> \<And>a. \<lbrace> \<lambda>s. I a s \<and> C a s \<rbrace> B a \<lbrace> I \<rbrace>;
    \<And>a s. \<lbrakk> I a s; \<not> C a s \<rbrakk> \<Longrightarrow> Q a s
   \<rbrakk> \<Longrightarrow> \<lbrace> I a \<rbrace> whileLoop C B a \<lbrace> Q \<rbrace>"
  apply (subst valid_def)
  apply (rule allI, rule impI)
  apply atomize
  apply (rule whileLoop_rule'[where Q="\<lambda>B. \<forall>(x, y) \<in> fst B. Q x y" and I="\<lambda>r s. I r s"])
    apply simp
   apply (clarsimp simp: valid_def, force)
  apply (clarsimp simp: valid_def, force)
  done

lemma whileLoop_wp_inv [wp]:
  "\<lbrakk> \<And>a. \<lbrace>\<lambda>s. I a s \<and> C a s\<rbrace> B a \<lbrace>I\<rbrace>; \<And>a s. \<lbrakk>I a s; \<not> C a s\<rbrakk> \<Longrightarrow> Q a s \<rbrakk>
      \<Longrightarrow> \<lbrace> I a \<rbrace> whileLoop_inv C B a I M \<lbrace> Q \<rbrace>"
  apply (clarsimp simp: whileLoop_inv_def)
  apply (rule whileLoop_wp, auto)
  done

(* Testcase for whileLoop_wp *)
lemma "\<lbrace> P \<rbrace> whileLoop (\<lambda>n _. n < (10 :: nat)) (\<lambda>n. return (n + 1)) 10 \<lbrace> \<lambda>x s. x = 10 \<rbrace>"
  apply (wp whileLoop_wp [where I="\<lambda>n _. n \<le> 10"])
  apply simp_all
  done

lemma whileLoopE_wp:
  "\<lbrakk> \<And>a. \<lbrace> \<lambda>s. I a s \<and> C a s \<rbrace> B a \<lbrace> I \<rbrace>,\<lbrace> A \<rbrace>;
    \<And>a s. \<lbrakk> I a s; \<not> C a s \<rbrakk> \<Longrightarrow> Q a s
   \<rbrakk> \<Longrightarrow> \<lbrace> I a \<rbrace> whileLoopE C B a \<lbrace> Q \<rbrace>,\<lbrace> A \<rbrace>"
  apply (clarsimp simp: whileLoopE_def validE_def)
  apply (rule hoare_chain)
    apply (rule whileLoop_wp [where I="\<lambda>r s. (case r of Inl x \<Rightarrow> A x s | Inr x \<Rightarrow> I x s)"
               and Q="\<lambda>r s. (case r of Inl x \<Rightarrow> A x s | Inr x \<Rightarrow> Q x s)"])
     apply atomize
     apply (clarsimp simp: valid_def lift_def split: sum.splits)
    apply (clarsimp simp: valid_def lift_def split: sum.splits)
   apply clarsimp
  apply clarsimp
  done

lemma whileLoop_terminatesE_induct [consumes 1]:
  assumes induct: "whileLoop_terminatesE C B r s"
  and init: "\<And>r s. \<not> C r s \<Longrightarrow> I r s"
  and step: "\<And>r s r' s'. \<lbrakk> C r s; \<forall>(v', s') \<in> fst (B r s). case v' of Inl _ \<Rightarrow> True | Inr r' \<Rightarrow> I r' s' \<rbrakk> \<Longrightarrow> I r s"
  shows "I r s"
  apply (insert induct)
  apply (clarsimp simp: whileLoop_terminatesE_def)
  apply (subgoal_tac "(\<lambda>r s. case (Inr r) of Inl x \<Rightarrow> True | Inr x \<Rightarrow> I x s) r s")
   apply simp
  apply (induction rule: whileLoop_terminates.induct)
   apply (case_tac r)
    apply simp
   apply clarsimp
   apply (erule init)
  apply (clarsimp split: sum.splits)
  apply (rule step)
   apply simp
  apply (clarsimp simp: lift_def split: sum.splits)
  apply force
  done

lemma whileLoop_terminates_unfold:
  "\<lbrakk> whileLoop_terminates C B r s; (r', s') \<in> fst (B r s); C r s \<rbrakk> \<Longrightarrow> whileLoop_terminates C B r' s'"
  apply (erule whileLoop_terminates.cases)
   apply simp
  apply force
  done

lemma whileLoop_terminatesE_unfold:
  "\<lbrakk> whileLoop_terminatesE C B r s; (Inr r', s') \<in> fst (B r s); C r s \<rbrakk> \<Longrightarrow> whileLoop_terminatesE C B r' s'"
  apply (clarsimp simp: whileLoop_terminatesE_def)
  apply (drule whileLoop_terminates_unfold)
    apply (clarsimp simp: lift_def)
    apply assumption
   apply clarsimp
  apply clarsimp
  done

lemma whileLoop_terminates_cond_false:
  "\<not> C r s \<Longrightarrow> whileLoop_terminates C B r s"
  apply (erule whileLoop_terminates.intros)
  done

lemma whileLoop_terminates_cond_true:
  "\<lbrakk> C r s; \<forall>(r', s') \<in> fst (B r s). whileLoop_terminates C B r' s' \<rbrakk> \<Longrightarrow> whileLoop_terminates C B r s"
  apply (erule (1) whileLoop_terminates.intros)
  done

lemma whileLoop_terminatesE_cond_false:
  "\<not> C r s \<Longrightarrow> whileLoop_terminatesE C B r s"
  apply (clarsimp simp: whileLoop_terminatesE_def)
  apply (rule whileLoop_terminates_cond_false)
  apply simp
  done

lemma whileLoop_terminatesE_cond_true:
  "\<lbrakk> C r s; \<forall>(v', s') \<in> fst (B r s). case v' of Inl _ \<Rightarrow> True | Inr r' \<Rightarrow> whileLoop_terminatesE C B r' s' \<rbrakk>
         \<Longrightarrow> whileLoop_terminatesE C B r s"
  apply (clarsimp simp: whileLoop_terminatesE_def)
  apply (rule whileLoop_terminates_cond_true)
   apply simp
  apply (clarsimp simp: lift_def)
  apply (case_tac a)
   apply (clarsimp simp: lift_def split: sum.splits)
   apply (rule whileLoop_terminates_cond_false)
   apply (clarsimp simp: lift_def)
  apply clarsimp
  apply (fastforce simp: whileLoop_terminatesE_def)
  done

lemma whileLoop_terminates_cong:
  assumes r: "r = r'"
  and s: "s = s'"
  and C: "\<And>r s. C r s = C' r s"
  and B: "\<And>r s. C' r s \<Longrightarrow> B r s = B' r s"
  shows "whileLoop_terminates C B r s = whileLoop_terminates C' B' r' s'"
proof (rule iffI)
  assume T: "whileLoop_terminates C B r s"
  show "whileLoop_terminates C' B' r' s'"
    apply (insert T r s)
    apply (induct arbitrary: r' s' rule: whileLoop_terminates.induct)
     apply (rule whileLoop_terminates_cond_false)
     apply (clarsimp simp: C)
    apply (rule whileLoop_terminates_cond_true)
     apply (clarsimp simp: C)
    apply (clarsimp simp: split_def B C)
    done
next
  assume T: "whileLoop_terminates C' B' r' s'"
  show "whileLoop_terminates C B r s"
    apply (insert T r s)
    apply (induct arbitrary: r s rule: whileLoop_terminates.induct)
     apply (rule whileLoop_terminates_cond_false)
     apply (clarsimp simp: C)
    apply (rule whileLoop_terminates_cond_true)
     apply (clarsimp simp: C)
    apply (clarsimp simp: split_def B C)
    done
qed

lemma whileLoop_terminatesE_cong:
  "\<lbrakk> r = r'; s = s'; \<And>r s. C r s = C' r s; \<And>r s. C' r s \<Longrightarrow> B r s = B' r s \<rbrakk> \<Longrightarrow> whileLoop_terminatesE C B r s = whileLoop_terminatesE C' B' r' s'"
  apply (clarsimp simp: whileLoop_terminatesE_def)
  apply (rule whileLoop_terminates_cong, simp_all)
   apply (clarsimp split: sum.splits)
  apply (clarsimp simp: NonDetMonad.lift_def throwError_def split: sum.splits)
  done

lemma whileLoop_cong:
  "\<lbrakk> \<And>r s. C r s = C' r s; i = i'; \<And>r s. C r s \<Longrightarrow> B r s = B' r s \<rbrakk> \<Longrightarrow> whileLoop C B i = whileLoop C' B' i'"
  apply (rule ext)
  apply (clarsimp simp: whileLoop_def cong: whileLoop_results_cong whileLoop_terminates_cong)
  done

lemma whileLoop_cong':
  "\<lbrakk> \<And>r s. C r s = C' r s; i = i'; \<And>r s. C r s \<Longrightarrow> B r s = B' r s; s = s' \<rbrakk>
        \<Longrightarrow> whileLoop C B i s = whileLoop C' B' i' s'"
  apply (clarsimp simp: whileLoop_def cong: whileLoop_results_cong whileLoop_terminates_cong)
  done

lemma whileLoopE_cong:
  "\<lbrakk> \<And>r s. C r s = C' r s ; i = i'; \<And>r s. C r s \<Longrightarrow> B r s = B' r s \<rbrakk> \<Longrightarrow> whileLoopE C B i = whileLoopE C' B' i'"
  apply (clarsimp simp: whileLoopE_def)
  apply (rule whileLoop_cong)
    apply (clarsimp split: sum.splits)
   apply (clarsimp split: sum.splits)
  apply (clarsimp simp: lift_def throwError_def split: sum.splits)
  done

lemma whileLoopE_cong':
  "\<lbrakk> \<And>r s. C r s = C' r s; i = i'; \<And>r s. C r s \<Longrightarrow> B r s = B' r s; s = s' \<rbrakk> \<Longrightarrow> whileLoopE C B i s = whileLoopE C' B' i' s'"
  apply (clarsimp cong: whileLoopE_cong)
  done

lemma snd_whileLoop_first_step: "\<lbrakk> \<not> snd (whileLoop C B r s); C r s \<rbrakk> \<Longrightarrow> \<not> snd (B r s)"
  apply (subst (asm) whileLoop_unroll)
  apply (clarsimp simp: bind_def condition_def)
  done

lemma snd_whileLoopE_first_step: "\<lbrakk> \<not> snd (whileLoopE C B r s); C r s \<rbrakk> \<Longrightarrow> \<not> snd (B r s)"
  apply (subgoal_tac "\<lbrakk> \<not> snd (whileLoopE C B r s); C r s \<rbrakk> \<Longrightarrow> \<not> snd ((lift B (Inr r)) s)")
   apply (clarsimp simp: lift_def)
  apply (unfold whileLoopE_def)
  apply (erule snd_whileLoop_first_step)
  apply clarsimp
  done

lemma snd_whileLoop_unfold:
      "\<lbrakk> \<not> snd (whileLoop C B r s); C r s; (r', s') \<in> fst (B r s) \<rbrakk> \<Longrightarrow> \<not> snd (whileLoop C B r' s')"
  apply (clarsimp simp: whileLoop_def)
  apply (auto simp: elim: whileLoop_results.cases whileLoop_terminates.cases
     intro: whileLoop_results.intros whileLoop_terminates.intros)
  done

lemma snd_whileLoopE_unfold:
      "\<lbrakk> \<not> snd (whileLoopE C B r s); (Inr r', s') \<in> fst (B r s); C r s \<rbrakk> \<Longrightarrow> \<not> snd (whileLoopE C B r' s')"
  apply (clarsimp simp: whileLoopE_def)
  apply (drule snd_whileLoop_unfold)
    apply clarsimp
   apply (clarsimp simp: lift_def)
   apply assumption
  apply (clarsimp simp: lift_def)
  done

lemma empty_fail_whileLoop:
  assumes body_empty_fail: "\<And>x. empty_fail (B x)"
  shows "empty_fail (whileLoop C B x)"
proof -
  {
    fix s A
    assume empty: "fst (whileLoop C B x s) = {}"

    have cond_true: "\<And>x s. fst (whileLoop C B x s) = {} \<Longrightarrow> C x s"
      apply (subst (asm) whileLoop_unroll)
      apply (clarsimp simp: condition_def return_def split: split_if_asm)
      done

    have  "snd (whileLoop C B x s)"
      apply (rule snd_whileLoop [where I="\<lambda>x s. fst (whileLoop C B x s) = {}"])
        apply fact
       apply (rule cond_true, fact)
      apply (case_tac "fst (B r s) = {}")
       apply (metis empty_failD [OF body_empty_fail])
      apply (subst (asm) whileLoop_unroll)
      apply (fastforce simp: condition_def bind_def split_def cond_true)
      done
   }

  thus ?thesis
    by (clarsimp simp: empty_fail_def)
qed

lemma empty_fail_whileLoopE:
  assumes body_empty_fail: "\<And>x. empty_fail (B x)"
  shows "empty_fail (whileLoopE C B x)"
  apply (clarsimp simp: whileLoopE_def)
  apply (rule empty_fail_whileLoop)
  apply (insert body_empty_fail)
  apply (clarsimp simp: empty_fail_def lift_def throwError_def return_def split: sum.splits)
  done

lemma whileLoop_results_bisim:
  assumes base: "(a, b) \<in> whileLoop_results C B"
  and vars1: "Q = (case a of Some (r, s) \<Rightarrow> Some (rt r, st s) | _ \<Rightarrow> None)"
  and vars2: "R = (case b of Some (r, s) \<Rightarrow> Some (rt r, st s) | _ \<Rightarrow> None)"
  and inv_init: "case a of Some (r, s) \<Rightarrow> I r s | _ \<Rightarrow> True"
  and inv_step: "\<And>r s r' s'. \<lbrakk> I r s; C r s; (r', s') \<in> fst (B r s) \<rbrakk> \<Longrightarrow> I r' s'"
  and cond_match: "\<And>r s. I r s \<Longrightarrow> C r s = C' (rt r) (st s)"
  and fail_step: "\<And>r s. \<lbrakk>C r s; snd (B r s); I r s\<rbrakk>
          \<Longrightarrow> (Some (rt r, st s), None) \<in> whileLoop_results C' B'"
  and refine: "\<And>r s r' s'. \<lbrakk> I r s; C r s; (r', s') \<in>  fst (B r s) \<rbrakk>
                           \<Longrightarrow> (rt r', st s') \<in> fst (B' (rt r) (st s))"
  shows "(Q, R) \<in> whileLoop_results C' B'"
  apply (subst vars1)
  apply (subst vars2)
  apply (insert base inv_init)
  apply (induct rule: whileLoop_results.induct)
    apply clarsimp
    apply (subst (asm) cond_match)
     apply (clarsimp simp: option.splits)
    apply (clarsimp simp: option.splits)
   apply (clarsimp simp: option.splits)
   apply (metis fail_step)
  apply (case_tac z)
   apply (clarsimp simp: option.splits)
  apply (metis cond_match inv_step refine whileLoop_results.intros(4))
   apply (clarsimp simp: option.splits)
 apply (metis cond_match inv_step refine whileLoop_results.intros(4))
 done

lemma whileLoop_terminates_liftE:
  "whileLoop_terminatesE C (\<lambda>r. liftE (B r)) r s = whileLoop_terminates C B r s"
  apply (subst eq_sym_conv)
  apply (clarsimp simp: whileLoop_terminatesE_def)
  apply (rule iffI)
   apply (erule whileLoop_terminates.induct)
    apply (rule whileLoop_terminates.intros)
    apply clarsimp
   apply (clarsimp simp: split_def)
   apply (rule whileLoop_terminates.intros(2))
    apply clarsimp
   apply (clarsimp simp: liftE_def in_bind return_def lift_def [abs_def]
               bind_def lift_def throwError_def split: sum.splits)
   apply (drule (1) bspec)
   apply (clarsimp)
  apply (subgoal_tac "case (Inr r) of Inl _ \<Rightarrow> True | Inr r \<Rightarrow>
      whileLoop_terminates (\<lambda>r s. (\<lambda>r s. case r of Inl _ \<Rightarrow> False | Inr v \<Rightarrow> C v s) (Inr r) s)
           (\<lambda>r. (lift (\<lambda>r. liftE (B r)) (Inr r)) >>= (\<lambda>x. return (theRight x))) r s")
   apply (clarsimp simp: liftE_def lift_def)
  apply (erule whileLoop_terminates.induct)
   apply (clarsimp simp: liftE_def lift_def split: sum.splits)
   apply (erule whileLoop_terminates.intros)
  apply (clarsimp simp: liftE_def split: sum.splits)
  apply (clarsimp simp: bind_def return_def split_def lift_def)
  apply (erule whileLoop_terminates.intros)
  apply force
  done

lemma snd_X_return [simp]: "\<And>A X s. snd ((A >>= (\<lambda>a. return (X a))) s) = snd (A s)"
  by (clarsimp simp: return_def bind_def split_def)

lemma whileLoopE_liftE:
  "whileLoopE c (\<lambda>r. liftE (B r)) i = liftE (whileLoop c B i)"
  apply (rule ext)
  apply (clarsimp simp: whileLoopE_def)
  apply (rule prod_eqI)
   apply (rule set_eqI, rule iffI)
    apply clarsimp
    apply (clarsimp simp: in_bind whileLoop_def liftE_def)
    apply (rule_tac x="b" in exI)
    apply (rule_tac x="theRight a" in exI)
    apply (rule conjI)
     apply (erule whileLoop_results_bisim [where rt=theRight and st="\<lambda>x. x" and I="\<lambda>r s. case r of Inr x \<Rightarrow> True | _ \<Rightarrow> False"],
    auto intro: whileLoop_results.intros intro!: bexI simp: bind_def return_def lift_def split: sum.splits)[1]
    apply (drule whileLoop_rule_lemma2 [where P="\<lambda>(r, s). case r of Inr x \<Rightarrow> True | _ \<Rightarrow> False"] )
        apply (rule refl)
       apply (rule refl)
      apply clarsimp
     apply (clarsimp simp: return_def bind_def lift_def split: sum.splits)
    apply (clarsimp simp: return_def bind_def lift_def split: sum.splits)
   apply (clarsimp simp: in_bind whileLoop_def liftE_def)
   apply (erule whileLoop_results_bisim [where rt=Inr and st="\<lambda>x. x" and I="\<lambda>r s. True"],
     auto intro: whileLoop_results.intros intro!: bexI simp: bind_def return_def lift_def split: sum.splits)[1]
  apply (rule iffI)
   apply (clarsimp simp: whileLoop_def liftE_def del: notI)
   apply (erule disjE)
    apply (erule whileLoop_results_bisim [where rt=theRight and st="\<lambda>x. x" and I="\<lambda>r s. case r of Inr x \<Rightarrow> True | _ \<Rightarrow> False"],
      auto intro: whileLoop_results.intros intro!: bexI simp: bind_def return_def lift_def split: sum.splits)[1]
   apply (subst (asm) whileLoop_terminates_liftE [symmetric])
   apply (fastforce simp: whileLoop_def liftE_def whileLoop_terminatesE_def)
  apply (clarsimp simp: whileLoop_def liftE_def del: notI)
  apply (subst (asm) whileLoop_terminates_liftE [symmetric])
  apply (clarsimp simp: whileLoop_def liftE_def whileLoop_terminatesE_def)
  apply (erule disjE)
   apply (erule whileLoop_results_bisim [where rt=Inr and st="\<lambda>x. x" and I="\<lambda>r s. True"])
         apply (clarsimp split: option.splits)
        apply (clarsimp split: option.splits)
       apply (clarsimp split: option.splits)
      apply (auto intro: whileLoop_results.intros intro!: bexI simp: bind_def return_def lift_def split: sum.splits)
  done

lemma whileLoop_results_induct_lemma':
  "\<lbrakk>(x1, x2) \<in> whileLoop_results C B; x1 = Some (r, s); x2 = Some (r', s');
    \<And>y. (x1, y) \<in> whileLoop_results C B \<Longrightarrow> \<exists>z. y = Some z;
    \<And>r s. \<not> C r s \<Longrightarrow> P r s r s;
    \<And>r s r' s' r'' s''. \<lbrakk>
        C r s;
       (r', s') \<in> fst (B r s);
       (Some (r', s'), Some (r'', s'')) \<in> whileLoop_results C B;
       \<And>y. (Some (r, s), y) \<in> whileLoop_results C B \<Longrightarrow> \<exists>z. y = Some z;
       P r' s' r'' s''
     \<rbrakk> \<Longrightarrow> P r s r'' s''
   \<rbrakk> \<Longrightarrow> P r s r' s'"
  apply (induct arbitrary: r s r' s' rule: whileLoop_results.induct)
     apply clarsimp
    apply clarsimp
   apply clarsimp
  apply (clarsimp simp: whileLoop_results.intros)
  done

lemma whileLoop_induct' [consumes 1]:
  assumes in_whileLoop: "(r', s') \<in> fst (whileLoop C B r s)"
  and no_fail: "\<not> snd (whileLoop C B r s)"
  and cond: "\<And> r s. \<not> C r s \<Longrightarrow> I r s r s"
   and induct: "\<And>r s r' s' r'' s''.
        \<lbrakk> C r s; (r', s') \<in> fst (B r s);
          (r'', s'') \<in> fst (whileLoop C B r' s');
          \<not> snd (whileLoop C B r s);
           I r' s' r'' s'' \<rbrakk> \<Longrightarrow> I r s r'' s''"
  shows "I r s r' s'"
  apply (insert in_whileLoop no_fail)
  apply (clarsimp simp: whileLoop_def)
  apply (erule rev_mp [where P="whileLoop_terminates C B r s"])
  apply (erule whileLoop_results_induct_lemma')
      apply simp
     apply simp
    apply (case_tac y)
     apply simp
    apply simp
   apply (clarsimp simp: cond)
  apply clarsimp
  apply (frule (2) whileLoop_terminates_unfold)
  apply (rule induct, auto simp: whileLoop_def)
  done

lemma whileLoopE_induct_lemma [consumes 1]:
  assumes induct: "(r', s') \<in> fst (whileLoop C' B' r s)"
  and new_cond: "C' = (\<lambda>r s. case r of Inl a \<Rightarrow> False | Inr v \<Rightarrow> C v s)"
  and new_body: "B' = (lift B)"
  and init_normal: "\<exists>x. r = Inr x"
  and init_case: "\<And>r s. \<not> C r s \<Longrightarrow> I (Inr r) s (Inr r) s"
  and step_case: "\<And>r s r' s' r'' s''. \<lbrakk> C r s; (Inr r', s') \<in> fst (B r s);  I (Inr r') s' r'' s'';  (r'', s'') \<in> fst (whileLoopE C B r' s') \<rbrakk> \<Longrightarrow> I (Inr r) s r'' s''"
  and except_case: "\<And>r s r' s' r'' s''. \<lbrakk> C r s; (Inl r', s') \<in> fst (B r s) \<rbrakk> \<Longrightarrow> I (Inr r) s (Inl r') s'"
  shows "I r s r' s'"
  apply (subgoal_tac "(case r of Inr x \<Rightarrow> I r s r' s' | Inl y \<Rightarrow> True)")
   apply (insert init_normal)[1]
   apply clarsimp
  apply (insert induct)
  apply (erule whileLoop_induct)
   apply (clarsimp simp: new_cond init_case split: sum.splits)
  apply (case_tac r')
   apply (clarsimp simp: new_cond new_body lift_def split: sum.splits)
   apply (drule (1) except_case)
   apply (subst (asm) whileLoop_unroll)
   apply (fastforce simp: return_def throwError_def split: sum.splits condition_splits cong: sum.case_cong)
  apply (clarsimp simp: new_cond new_body lift_def split: sum.splits)
  apply (drule (2) step_case)
   apply (clarsimp simp: whileLoopE_def lift_def [abs_def]  split: sum.splits cong: sum.case_cong)
  apply simp
  done

lemma whileLoopE_induct [consumes 1]:
  assumes induct: "(r' :: ('a + 'b), s') \<in> fst (whileLoopE C B r s)"
  and v: "Inr r = (v :: ('a + 'b))"
  and init_case: "\<And>r s. \<not> C r s \<Longrightarrow> I (Inr r) s (Inr r) s"
  and step_case: "\<And>r s r' s' r'' s''. \<lbrakk> C r s; (Inr r', s') \<in> fst (B r s);  I (Inr r') s' r'' s'';  (r'', s'') \<in> fst (whileLoopE C B r' s') \<rbrakk> \<Longrightarrow> I (Inr r) s r'' s''"
  and except_case: "\<And>r s r' s' r'' s''. \<lbrakk> C r s; (Inl r', s') \<in> fst (B r s) \<rbrakk> \<Longrightarrow> I (Inr r) s (Inl r') s'"
  shows "I v s r' s'"
  apply (insert induct)
  apply (clarsimp simp: whileLoopE_def)
  apply (subst (asm) v | subst v)+
  apply (erule whileLoopE_induct_lemma, (rule refl)+)
     apply (insert v)[1]
     apply force
    apply (erule init_case)
   apply (erule step_case, assumption+)
  apply (erule except_case, assumption+)
  done

lemma snd_whileLoop_induct_lemma [consumes 1]:
  assumes induct: "(x, y) \<in> whileLoop_results C B"
  and a: "x = Some (r, s)"
  and b: "y = None"
  and init: "\<And>r s. \<lbrakk> snd (B r s); C r s \<rbrakk> \<Longrightarrow> I r s"
  and step: "\<And>r s r' s'. \<lbrakk>C r s; (r', s') \<in> fst (B r s); (Some (r', s'), None) \<in> whileLoop_results C B; I r' s'\<rbrakk> \<Longrightarrow> I r s"
  shows "I r s"
  apply (insert induct init a b)
  apply atomize
  apply (induct arbitrary: r s rule: whileLoop_results.induct)
    apply (auto elim!: step)
  done

lemma snd_whileLoop_induct [consumes 1]:
  assumes induct: "snd (whileLoop C B r s)"
  and terminates: "\<not> whileLoop_terminates C B r s \<Longrightarrow> I r s"
  and init: "\<And> r s. \<lbrakk> snd (B r s); C r s \<rbrakk> \<Longrightarrow> I r s"
  and step: "\<And>r s r' s' r'' s''.
        \<lbrakk> C r s; (r', s') \<in> fst (B r s);
           snd (whileLoop C B r' s');
           I r' s' \<rbrakk> \<Longrightarrow> I r s"
  shows "I r s"
  apply (insert init induct)
  apply atomize
  apply (unfold whileLoop_def)
  apply clarsimp
  apply (erule disjE)
   apply (erule rev_mp)
   apply (erule snd_whileLoop_induct_lemma)
      apply simp
     apply simp
    apply clarsimp
   apply clarsimp
   apply (erule (1) step)
    apply (clarsimp simp: whileLoop_def)
   apply clarsimp
  apply (metis terminates)
  done

lemma snd_whileLoopE_induct [consumes 1]:
  assumes induct: "snd (whileLoopE C B r s)"
  and terminates: "\<not> whileLoop_terminatesE C B r s \<Longrightarrow> I r s"
  and init: "\<And> r s. \<lbrakk> snd (B r s); C r s \<rbrakk> \<Longrightarrow> I r s"
  and step: "\<And>r s r' s' r'' s''.
        \<lbrakk> C r s; (Inr r', s') \<in> fst (B r s);
           snd (whileLoopE C B r' s');
           I r' s' \<rbrakk> \<Longrightarrow> I r s"
  shows "I r s"
  apply (subgoal_tac "I (theRight (Inr r)) s")
   apply simp
  apply (insert init induct)
  apply atomize
  apply (erule rev_mp)
  apply (unfold whileLoopE_def)
  apply (erule snd_whileLoop_induct)
    apply clarsimp
    apply (rule terminates)
    apply (clarsimp simp: whileLoop_terminatesE_def)
   apply (clarsimp simp: lift_def throwError_def return_def split: sum.splits)
  apply (clarsimp split: sum.splits)
  apply (case_tac r')
   apply clarsimp
   apply (subst (asm) whileLoop_unroll)
   apply (fastforce split: condition_splits simp: return_def)
  apply (clarsimp simp: lift_def)
  apply (erule (1) step)
   apply (clarsimp simp: whileLoopE_def split: sum.splits)
  apply simp
  done

lemma no_fail_whileLoop:
  assumes PI: "\<And>s. P s \<Longrightarrow> I a s"
  and BI: "\<And>a. \<lbrace>\<lambda>s. I a s \<and> C a s\<rbrace> B a \<lbrace>I\<rbrace>"
  and NF: "\<And>a. no_fail (\<lambda>s. I a s \<and> C a s) (B a)"
  and LT: "\<And>s. P s \<Longrightarrow> whileLoop_terminates C B a s"
  shows "no_fail P (whileLoop C B a)"
  apply (clarsimp simp: no_fail_def)
  apply (cut_tac s=s in PI)
   apply clarsimp
  apply (erule_tac P="I a s" in rev_mp)
  apply (erule snd_whileLoop_induct)
    apply (metis LT)
   apply (cut_tac a=r in NF)
   apply (clarsimp simp: no_fail_def)
  apply clarsimp
  apply (cut_tac a=r in BI)
  apply (clarsimp simp: valid_def)
  apply (erule_tac x=sa in allE)
  apply force
  done

lemma no_fail_whileLoopE:
  assumes PI: "\<And>s. P s \<Longrightarrow> I a s"
  and BI: "\<And>a. \<lbrace>\<lambda>s. I a s \<and> C a s\<rbrace> B a \<lbrace>I\<rbrace>,\<lbrace>\<lambda>r s. True \<rbrace>"
  and NF: "\<And>a. no_fail (\<lambda>s. I a s \<and> C a s) (B a)"
  and LT: "\<And>s. P s \<Longrightarrow> whileLoop_terminatesE C B a s"
  shows "no_fail P (whileLoopE C B a)"
  apply (insert PI BI NF LT)
  apply (clarsimp simp: whileLoopE_def whileLoop_terminatesE_def validE_def)
  apply (rule no_fail_whileLoop [where I="\<lambda>r s. case r of Inr x \<Rightarrow> I x s | _ \<Rightarrow> True"])
     apply clarsimp
    apply (clarsimp simp: NonDetMonad.lift_def split: sum.splits)
   apply (clarsimp simp: NonDetMonad.lift_def no_fail_def split: sum.splits option.splits)
  apply (clarsimp simp: NonDetMonad.lift_def sum.splits)
  done

lemma whileLoop_terminates_wf_measure:
  assumes inv_start: "I a s"
  and inv_body: "\<And>a. \<lbrace>\<lambda>s. I a s \<and> C a s\<rbrace> B a \<lbrace>I\<rbrace>"
  and measure: "\<And>a m. \<lbrace>\<lambda>s. I a s \<and> C a s \<and> m = (a, s) \<rbrace> B a \<lbrace> \<lambda>rv s. ((rv, s), m) \<in> D \<rbrace>"
  and wf_D: "wf D"
  shows "whileLoop_terminates C B a s"
  apply (insert inv_start)
  using wf_D
  apply (induct "(a, s)" arbitrary: a s)
  apply (case_tac "C a s")
   apply (rule whileLoop_terminates.intros, assumption)
   apply clarsimp
   apply (cut_tac a=a and m="(a, s)" in measure)
   apply (clarsimp simp: valid_def)
   apply (drule (1) bspec)
   apply atomize
   apply (metis use_valid inv_body)
  apply (rule whileLoop_terminates.intros)
  apply simp
  done

lemma wf_wellorder: "wf {(a, b). a < (b::'a::wellorder)}"
  apply (rule wfUNIVI)
  apply clarsimp
  apply (rule less_induct)
  apply atomize
  apply blast
  done

lemma wf_wellorder_measure: "wf {(a, b). (M a :: 'a :: wellorder) < M b}"
  apply (subgoal_tac "wf (inv_image ({(a, b). a < b}) M)")
   apply (clarsimp simp: inv_image_def)
  apply (rule wf_inv_image)
  apply (rule wf_wellorder)
  done

lemma whileLoop_terminates_measure:
  assumes inv_start: "I a s"
  and inv_body: "\<And>a. \<lbrace>\<lambda>s. I a s \<and> C a s\<rbrace> B a \<lbrace>I\<rbrace>"
  and measure: "\<And>a m. \<lbrace>\<lambda>s. I a s \<and> C a s \<and> M a s = m \<rbrace> B a \<lbrace> \<lambda>rv s. M rv s < (m :: 'a :: wellorder) \<rbrace>"
  shows "whileLoop_terminates C B a s"
  apply (rule whileLoop_terminates_wf_measure [where I=I and D="{((r, s), (r', s')). M r s < M r' s'}"])
     apply (rule inv_start)
    apply (rule inv_body)
   apply (insert measure)[1]
   apply (fastforce simp: valid_def)
  apply (clarsimp simp: split_def)
  apply (rule wf_wellorder_measure [unfolded split_def])
  done

lemma whileLoop_terminatesE_wf_measure:
  assumes inv_start: "I a s"
  and inv_body: "\<And>a. \<lbrace>\<lambda>s. I a s \<and> C a s\<rbrace> B a \<lbrace>I\<rbrace>,\<lbrace>\<lambda>r s. True \<rbrace>"
  and measure: "\<And>a m. \<lbrace>\<lambda>s. I a s \<and> C a s \<and> m = (a, s) \<rbrace> B a \<lbrace> \<lambda>rv s. ((rv, s), m) \<in> D \<rbrace>,\<lbrace>\<lambda>r s. True \<rbrace>"
  and wf_D: "wf D"
  shows "whileLoop_terminatesE C B a s"
  apply (insert inv_start inv_body measure wf_D)
  apply (clarsimp simp: whileLoop_terminatesE_def validE_def)
  apply (rule whileLoop_terminates_wf_measure [where I="\<lambda>r s. case r of Inr x \<Rightarrow> I x s | _ \<Rightarrow> True"
           and D="{((r', s'),(r, s)).
                   case r of Inr v \<Rightarrow> (
                                case r' of Inr v' \<Rightarrow> ((v', s'), (v, s)) \<in> D
                                         | Inl _ \<Rightarrow> True)
                         | _ \<Rightarrow> False}"])
    apply clarsimp
   apply (clarsimp simp: NonDetMonad.lift_def split: sum.splits)
  apply (clarsimp simp: NonDetMonad.lift_def split: sum.splits)
  apply (clarsimp simp: valid_def split: sum.splits, fastforce)
  apply (drule wf_map_pair_image [where f="\<lambda>(r, s). (Inr r, s)"])
   apply (rule injI, fastforce)
  apply (rule wf_subset,
     rule wf_union_compatible [where R="{((r', s'), (r, s)). (\<exists>x. r = Inr x) \<and> (\<exists>x. r' = Inl x)}"])
     apply (rule wf_no_loop)
     apply blast
    apply assumption
   apply (clarsimp simp: map_pair_def)
  apply atomize
  apply (rule subsetI)
  apply (clarsimp simp: map_pair_def image_def split_def split: sum.splits)
  apply (erule bexI [rotated], clarsimp)
  done

lemma whileLoop_terminatesE_measure:
  assumes inv_start: "I a s"
  and inv_body: "\<And>a. \<lbrace>\<lambda>s. I a s \<and> C a s\<rbrace> B a \<lbrace>I\<rbrace>,\<lbrace>\<lambda>r s. True \<rbrace> "
  and measure: "\<And>a m. \<lbrace>\<lambda>s. I a s \<and> C a s \<and> M a s = m \<rbrace> B a \<lbrace> \<lambda>rv s. M rv s < (m :: 'm :: wellorder) \<rbrace>,\<lbrace>\<lambda>r s. True \<rbrace>"
  shows "whileLoop_terminatesE C B a s"
  apply (rule whileLoop_terminatesE_wf_measure [where I=I and D="{((r, s), (r', s')). M r s < M r' s'}"])
     apply (rule inv_start)
    apply (rule inv_body)
   apply (insert measure)[1]
   apply (fastforce simp: validE_def2)
  apply (clarsimp simp: split_def)
  apply (rule wf_wellorder_measure [unfolded split_def])
  done

lemma snd_whileLoop_subset':
    assumes a_fails: "snd (whileLoop C A i s)"
    and b_success_step:
        "\<And>r s r' s'. \<lbrakk> C r s; (r', s') \<in> fst (A r s); \<not> snd (B r s) \<rbrakk>
             \<Longrightarrow> (r', s') \<in> fst (B r s)"
    and b_fail_step: "\<And>r s. \<lbrakk> C r s; snd (A r s) \<rbrakk> \<Longrightarrow> snd (B r s) "
   shows "snd (whileLoop C B i s)"
  apply (insert a_fails)
  apply (induct rule: snd_whileLoop_induct)
    apply (unfold whileLoop_def snd_conv)[1]
    apply (rule disjCI, simp)
    apply rotate_tac
    apply (induct rule: whileLoop_terminates.induct)
     apply (subst (asm) whileLoop_terminates.simps)
     apply simp
    apply (subst (asm) (3) whileLoop_terminates.simps, clarsimp)
    apply (subst whileLoop_results.simps, clarsimp)
    apply (rule classical)
    apply (frule b_success_step, assumption, simp)
    apply (drule (1) bspec)
    apply clarsimp
   apply (frule (1) b_fail_step)
   apply (metis snd_whileLoop_first_step)
  apply (metis b_success_step snd_whileLoop_first_step snd_whileLoop_unfold)
  done

lemma snd_whileLoop_subset:
    "\<lbrakk> snd (whileLoop C A i s);
         \<And>r s. \<not> snd (B r s) \<Longrightarrow> fst (A r s)  \<subseteq> fst (B r s);
         \<And>r s. snd (A r s) \<longrightarrow> snd (B r s) \<rbrakk>
   \<Longrightarrow> snd (whileLoop C B i s)"
  apply (erule snd_whileLoop_subset')
   apply fastforce
  apply fastforce
  done

lemma snd_whileLoopE_subset:
    "\<lbrakk> snd (whileLoopE C A i s);
         \<And>r s. \<not> snd (B r s) \<Longrightarrow> fst (A r s)  \<subseteq> fst (B r s);
         \<And>r s. snd (A r s) \<longrightarrow> snd (B r s) \<rbrakk>
   \<Longrightarrow> snd (whileLoopE C B i s)"
  apply (unfold whileLoopE_def)
  apply (erule snd_whileLoop_subset)
   apply (clarsimp simp: NonDetMonad.lift_def split: sum.splits)
  apply (clarsimp simp: NonDetMonad.lift_def split: sum.splits)
  done

lemma whileLoop_rule_strong:
  assumes init_U: "\<And>r'' s''. I r'' s'' r s"
      and iter_U:  "\<And>r'' s'' r. \<lbrace> \<lambda>s. I r'' s'' r s \<and> C r s \<rbrace> B r \<lbrace> \<lambda>r s. I r'' s'' r s \<rbrace>"
      and iter_L: "\<And>r s r'' s''. \<lbrakk> I r'' s'' r s; C r s\<rbrakk>
                    \<Longrightarrow> \<exists>(r', s')\<in>fst (B r s). I r'' s'' r' s' \<and> ((r', s'), r, s) \<in> R r'' s''"
      and wf_R: "\<And>r'' s''. \<lbrakk> I r'' s'' r'' s''; \<not> C r'' s'' \<rbrakk> \<Longrightarrow> wf (R r'' s'')"
      and end_Q:  "\<And>r'' s'' r s. \<lbrakk>I r'' s'' r s; \<not> C r'' s''; \<not> C r s\<rbrakk> \<Longrightarrow> r'' = r \<and> s'' = s"
      and Q_final: "Q ({(r'', s''). I r'' s'' r'' s'' \<and> \<not> C r'' s''}, snd (whileLoop C B r s))"
      shows "Q (whileLoop C B r s)"
proof -
  def fst_Q \<equiv> "{(r'', s''). I r'' s'' r'' s'' \<and> \<not> C r'' s''}"

  have lower_bound: "fst_Q \<subseteq> fst (whileLoop C B r s)"
    unfolding fst_Q_def
    apply (rule fst_whileLoop_lower_bound [where R=R and I=I])
        apply (clarsimp)
       apply (rule wf_R)
             apply clarsimp
           apply clarsimp
      apply (rule init_U)
     apply (rule iter_L, simp, simp)
    apply (rule end_Q, auto)
    done

  have upper_bound: "fst (whileLoop C B r s) \<subseteq> fst_Q"
    unfolding fst_Q_def
    apply (rule whileLoop_rule_weak [where I="\<lambda>r' s'. \<forall>r'' s''. I r'' s'' r' s'"])
      apply (clarsimp simp: init_U)
     apply (insert iter_U)[1]
     apply (clarsimp simp: valid_def)
     apply atomize
     apply blast
    apply clarsimp
    apply (drule (1) bspec)
    apply clarsimp
    done

  have "fst (whileLoop C B r s) = fst_Q"
    by (metis antisym_conv lower_bound upper_bound)

  thus ?thesis
    by (metis (full_types) Q_final fst_Q_def surjective_pairing)
qed


lemma infinite_whileLoop:
  assumes empty_fail: "\<And>r. empty_fail (X r)"
  shows  "((whileLoop (\<lambda>_ _. C) X r) = assert (\<not> C))"
     (is "?LHS = ?RHS")
  apply (case_tac "C")
   apply (rule ext)
   apply (rule prod_eqI)
    apply (rule whileLoop_rule_weak [where I="\<lambda>_ _. True"])
       apply simp
      apply simp
      apply (rule hoare_TrueI)
     apply (clarsimp simp: assert_def)
    apply (clarsimp simp: assert_def fail_def)
    apply force
   apply (clarsimp simp: assert_def fail_def)
   apply (rule snd_whileLoop [where I="\<lambda>_ _. True"])
     apply simp
    apply simp
   apply (insert empty_fail)
   apply (fastforce simp: empty_fail_def)
  apply (subst whileLoop_unroll)
  apply clarsimp
  done

end
